"""X10 devices."""
import asyncio

from ..constants import ResponseStatus, X10Commands
from ..events import OFF_EVENT, ON_EVENT, Event
from ..groups import DIMMABLE_LIGHT, ON_OFF_SWITCH
from ..groups.on_level import OnLevel
from ..groups.on_off import OnOff
from ..handlers.to_device.x10_send import X10CommandSend
from ..managers.x10_manager import (
    X10AllLightsOnOffManager,
    X10AllUnitsOffManager,
    X10DimBrightenManager,
    X10OnOffManager,
)
from ..utils import multiple_status
from .x10_base import X10DeviceBase

ALL_LIGHTS_ON_OFF = "all_lights_on_off"
ALL_UNITS_OFF = "all_units_off"
DIM_BRIGHT = "dim_bright"


class X10OnOffSensor(X10DeviceBase):
    """X10 On/Off device."""

    def __init__(self, housecode, unitcode):
        """Init the X10OnOffSensor class."""
        super().__init__(housecode, unitcode)
        self._description = "X10 On/Off Sensor"

    def _register_handlers_and_managers(self):
        self._managers[ON_OFF_SWITCH] = X10OnOffManager(self._address)
        self._managers[ALL_UNITS_OFF] = X10AllUnitsOffManager(self._address)

    def _register_groups(self):
        self._groups[1] = OnOff(ON_OFF_SWITCH, self._address, 1)

    def _register_events(self):
        self._events[1] = {}
        self._events[1][ON_EVENT] = Event(ON_EVENT, self._address, group=1)
        self._events[1][OFF_EVENT] = Event(OFF_EVENT, self._address, group=1)

    def _subscribe_to_handelers_and_managers(self):
        self._managers[ON_OFF_SWITCH].subscribe(self._groups[1].set_value)
        self._managers[ALL_UNITS_OFF].subscribe(self._groups[1].set_value)

        self._managers[ON_OFF_SWITCH].subscribe_on(self._events[1][ON_EVENT].trigger)
        self._managers[ON_OFF_SWITCH].subscribe_off(self._events[1][OFF_EVENT].trigger)
        self._managers[ALL_UNITS_OFF].subscribe(self._events[1][OFF_EVENT].trigger)


class X10OnOff(X10OnOffSensor):
    """X10 On/Off device."""

    def __init__(self, housecode, unitcode):
        """Init the X10OnOff class."""
        super().__init__(housecode, unitcode)
        self._description = "X10 On/Off Switch"
        self._on_cmd = X10CommandSend(self._address, X10Commands.ON)
        self._off_cmd = X10CommandSend(self._address, X10Commands.OFF)

    def on(self, group=0):
        """Turn on the switch."""
        asyncio.ensure_future(self.async_on())

    async def async_on(self, group=0):
        """Turn on the switch."""
        retries = 0
        while retries < 3:
            result = await self._on_cmd.async_send()
            if result == ResponseStatus.SUCCESS:
                self._groups[1].set_value(0xFF)
                return result
            retries += 1
        return ResponseStatus.FAILURE

    def off(self, group=0):
        """Turn off the switch."""
        asyncio.ensure_future(self.async_off())

    async def async_off(self, group=0):
        """Turn off the switch."""
        retries = 0
        while retries < 3:
            result = await self._off_cmd.async_send()
            if result == ResponseStatus.SUCCESS:
                self._groups[1].set_value(0x00)
                return result
            retries += 1
        return ResponseStatus.FAILURE

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._managers[ALL_LIGHTS_ON_OFF] = X10AllLightsOnOffManager(self._address)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()

        self._managers[ALL_LIGHTS_ON_OFF].subscribe(self._groups[1].set_value)
        self._managers[ALL_LIGHTS_ON_OFF].subscribe_on(
            self._events[1][ON_EVENT].trigger
        )
        self._managers[ALL_LIGHTS_ON_OFF].subscribe_off(
            self._events[1][OFF_EVENT].trigger
        )


class X10Dimmable(X10OnOff):
    """X10 Dimmable Switch."""

    def __init__(self, housecode, unitcode, steps: int = 22, max_level: int = 0xFF):
        """Init the X10Dimmable class."""
        super().__init__(housecode, unitcode)
        self._description = "X10 Dimmable Switch"
        self._steps = steps
        self._max_level = max_level
        self._increment = self._max_level / self._steps
        self._dim_cmd = X10CommandSend(self._address, X10Commands.DIM)
        self._bright_cmd = X10CommandSend(self._address, X10Commands.BRIGHT)

    @property
    def steps(self):
        """Return the number of steps from off to full on."""
        return self._steps

    # pylint: disable=arguments-renamed
    def on(self, on_level: int = 0xFF, group: int = 0):
        """Set the on level of the switch."""
        asyncio.ensure_future(self.async_on(on_level))

    # pylint: disable=arguments-renamed
    async def async_on(self, on_level: int = 0xFF, group: int = 0):
        """Set the on level of the switch."""
        if on_level == 0x00:
            return await self.async_off()
        if on_level == 0xFF:
            return await super().async_on()

        if on_level < 1:
            # Assume the user entered a precent on level
            on_level = on_level * self._max_level
        if self._groups[1].value is None:
            self._groups[1].set_value(0)
        change = on_level - self._groups[1].value
        steps = round(abs(change) / self._increment)
        method = self.async_bright if change > 0 else self.async_dim
        results = []
        for _ in range(0, steps):
            result = await method()
            results.append(result)
        return multiple_status(*results)

    def dim(self):
        """Dim the swich one step."""
        asyncio.ensure_future(self.async_dim)

    async def async_dim(self):
        """Dim the swich one step."""
        retries = 0
        while retries < 3:
            results = await self._dim_cmd.async_send()
            if results == ResponseStatus.SUCCESS:
                self._handle_dim_bright(-1)
                return results
            retries += 1
        return ResponseStatus.FAILURE

    def bright(self):
        """Dim the swich one step."""
        asyncio.ensure_future(self.async_bright)

    async def async_bright(self):
        """Dim the swich one step."""
        retries = 0
        while retries < 3:
            results = await self._bright_cmd.async_send()
            if results == ResponseStatus.SUCCESS:
                self._handle_dim_bright(1)
                return results
            retries += 1
        return ResponseStatus.FAILURE

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._managers[DIM_BRIGHT] = X10DimBrightenManager(self._address)

    def _register_groups(self):
        self._groups[1] = OnLevel(DIMMABLE_LIGHT, self._address, 1, 0)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._managers[DIM_BRIGHT].subscribe(self._handle_dim_bright)

    def _handle_dim_bright(self, on_level):
        """Handle a dim or bright command from the device."""
        new_value = self._groups[1].value + self._increment * on_level
        new_value = min(max(new_value, 0x00), 0xFF)
        self._groups[1].set_value(new_value)



"""Manage links beteween devices."""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, List, Union

import async_timeout

from ...address import Address
from ...aldb.aldb_record import ALDBRecord
from ...constants import AllLinkMode, EngineVersion, LinkStatus, ResponseStatus
from ...handlers.cancel_all_linking import CancelAllLinkingCommandHandler
from ...handlers.start_all_linking import StartAllLinkingCommandHandler
from ...handlers.to_device.engine_version_request import EngineVersionRequest
from ...handlers.to_device.enter_linking_mode import EnterLinkingModeCommand
from ...handlers.to_device.enter_unlinking_mode import EnterUnlinkingModeCommand

if TYPE_CHECKING:
    from ...device_types.device_base import Device

TIMEOUT = 3
_LOGGER = logging.getLogger(__name__)
_engine_version_queue = asyncio.Queue()


async def _handle_engine_version(engine_version: EngineVersion):
    """Handle the engine version response."""
    await _engine_version_queue.put(engine_version)


async def _get_engine_version(address: Address):
    """Get the engine version of the device."""

    while not _engine_version_queue.empty():
        _engine_version_queue.get_nowait()

    address = Address(address)
    cmd = EngineVersionRequest(address=address)
    cmd.subscribe(_handle_engine_version)
    retries = 3
    response = None
    while (
        response
        not in [
            ResponseStatus.SUCCESS,
            ResponseStatus.DIRECT_NAK_ALDB,
            ResponseStatus.DIRECT_NAK_PRE_NAK,
        ]
        and retries
    ):
        response = await cmd.async_send()
        retries -= 1
    if response == ResponseStatus.SUCCESS:
        try:
            async with async_timeout.timeout(2):
                return await _engine_version_queue.get()
        except asyncio.TimeoutError:
            pass
    elif response in [
        ResponseStatus.DIRECT_NAK_ALDB,
        ResponseStatus.DIRECT_NAK_PRE_NAK,
    ]:
        return EngineVersion.I2CS
    return EngineVersion.UNKNOWN


async def async_enter_linking_mode(
    link_mode: AllLinkMode, group: int, address: Address = None
):
    """Put the Insteon Modem into linking mode."""
    if address is not None:
        # pylint: disable=import-outside-toplevel
        from ... import devices

        address = Address(address)
        device = devices[address]
        if device:
            engine_version = device.engine_version
        else:
            engine_version = await _get_engine_version(address)

    link_cmd = StartAllLinkingCommandHandler()
    response_modem = await link_cmd.async_send(link_mode=link_mode, group=group)
    _LOGGER.debug("Enter linking mode modem response: %s", str(response_modem))

    response_device = ResponseStatus.UNSENT
    if address is not None:
        extended = engine_version in [EngineVersion.I2CS, EngineVersion.UNKNOWN]
        device_link_cmd = EnterLinkingModeCommand(address)
        retries = 3
        while retries and response_device != ResponseStatus.SUCCESS:
            response_device = await device_link_cmd.async_send(
                group=group, extended=extended
            )
            retries -= 1
        retries = 3
        while retries and response_device != ResponseStatus.SUCCESS:
            response_device = await device_link_cmd.async_send(
                group=group, extended=not extended
            )
            retries -= 1
        _LOGGER.debug("Enter linking mode device response: %s", str(response_device))
    return response_modem, response_device


async def async_enter_unlinking_mode(group: int, address: Address = None):
    """Put the Insteon Modem into unlinking mode."""

    if address is not None:
        # pylint: disable=import-outside-toplevel
        from ... import devices

        address = Address(address)
        device = devices[address]
        if device:
            engine_version = device.engine_version
        else:
            engine_version = await _get_engine_version(address)
    modem_link_cmd = StartAllLinkingCommandHandler()
    link_mode = AllLinkMode.DELETE
    response_modem = await modem_link_cmd.async_send(link_mode=link_mode, group=group)

    _LOGGER.debug("Enter linking mode modem response: %s", str(response_modem))

    response_device = ResponseStatus.UNSENT
    if address is not None:
        extended = engine_version in [EngineVersion.I2CS, EngineVersion.UNKNOWN]
        device_link_cmd = EnterUnlinkingModeCommand(address)
        retries = 3
        while retries and response_device != ResponseStatus.SUCCESS:
            response_device = await device_link_cmd.async_send(
                group=group, extended=extended
            )
            retries -= 1
        retries = 3
        while retries and response_device != ResponseStatus.SUCCESS:
            response_device = await device_link_cmd.async_send(
                group=group, extended=not extended
            )
            retries -= 1
        _LOGGER.debug("Enter linking mode device response: %s", str(response_device))

    return response_modem, response_device


async def async_link_devices(
    controller: Device, responder: Device, group: int = 0, data1=255, data2=28, data3=1
):
    """Link two devices."""

    failed_1 = 0
    failed_2 = 0

    if not hasattr(controller, "address") or not hasattr(responder, "address"):
        raise TypeError("controller and responder must be devices")
    responder.aldb.add(
        group=group,
        target=controller.address,
        controller=False,
        data1=data1,
        data2=data2,
        data3=data3,
    )
    controller.aldb.add(
        group=group,
        target=responder.address,
        controller=True,
        data1=data1,
        data2=data2,
        data3=data3,
    )
    _, failed_2 = await responder.aldb.async_write()
    _, failed_1 = await controller.aldb.async_write()

    if failed_1 or failed_2:
        return ResponseStatus.FAILURE
    return ResponseStatus.SUCCESS


async def async_cancel_linking_mode():
    """Cancel an All-Link session with the modem."""
    cmd = CancelAllLinkingCommandHandler()
    return await cmd.async_send()


async def async_unlink_devices(
    device1: Device, device2: Union[Device, Address], group: Union(int, None) = None
):
    """Unlink two devices.

    `device2` can be passed as `Device` or `Address`. When `device2` are passed as `Address`, only the links in `device1` will be removed.

    `group` can be passed as an `int` or `None`. When `group` is `None` all links between the two devices will be removed. When
    `group` is an `int`, only that group wil be removed.
    """
    try:
        addr1 = device1.address
    except AttributeError as ex:
        raise TypeError("device1 must be a Device") from ex

    device2_is_device = True
    try:
        addr2 = device2.address
    except AttributeError:
        try:
            addr2 = Address(device2)
            device2_is_device = False
        except TypeError as ex:
            raise TypeError("device2 must be a Device or an Address") from ex

    failed_1 = 0
    failed_2 = 0

    for rec in device1.aldb.find(target=addr2, group=group):
        device1.aldb.modify(mem_addr=rec.mem_addr, in_use=False)
        _, failed_1 = await device1.aldb.async_write()

    if device2_is_device:
        for rec in device2.aldb.find(target=addr1, group=group):
            device2.aldb.modify(mem_addr=rec.mem_addr, in_use=False)
            _, failed_2 = await device2.aldb.async_write()

    if failed_1 or failed_2:
        return ResponseStatus.FAILURE
    return ResponseStatus.SUCCESS


def find_broken_links(devices: List[Device]):
    """Find proken links."""
    broken_link_list = {}
    for addr in devices:
        device = devices[addr]
        for mem_addr in device.aldb:
            rec = device.aldb[mem_addr]
            if rec.is_in_use:
                status = _test_broken(addr, rec, devices)
                if status != LinkStatus.FOUND:
                    if not broken_link_list.get(addr):
                        broken_link_list[addr] = {}
                    broken_link_list[addr][mem_addr] = (rec, status)
    return broken_link_list


def _test_broken(address: Address, rec: ALDBRecord, devices: List[Device]):
    """Test if a corresponding record exists in liked device."""
    device = devices.get(rec.target)
    if not device:
        return LinkStatus.MISSING_TARGET

    if not device.aldb.is_loaded:
        return LinkStatus.TARGET_DB_NOT_LOADED

    for mem_addr in device.aldb:
        t_rec = device.aldb[mem_addr]
        if (
            t_rec.target == address
            and t_rec.group == rec.group
            and t_rec.is_controller != rec.is_controller
        ):
            return LinkStatus.FOUND
    if rec.is_controller:
        return LinkStatus.MISSING_RESPONDER
    return LinkStatus.MISSING_CONTROLLER



"""Send a Start All-Linking command."""
from .send_all_link import SendAllLinkCommandHandler


class SendAllLinkOnCommandHandler(SendAllLinkCommandHandler):
    """Send Start All-LInking ON commands."""

    # pylint: disable=arguments-differ
    async def async_send(self, group: int):
        """Send the Start All-Linking message asyncronously."""
        return await super().async_send(group=group, cmd1=0x11, cmd2=0x00)



"""Thermostat status manager."""

import asyncio
from collections import namedtuple
import logging

from ..address import Address
from ..constants import ResponseStatus, ThermostatMode
from ..handlers.from_device.ext_get_response import ExtendedGetResponseHandler
from ..handlers.to_device.extended_get import ExtendedGetCommand
from ..utils import (
    bit_is_set,
    calc_thermostat_mode,
    calc_thermostat_temp,
    multiple_status,
    publish_topic,
    subscribe_topic,
    unsubscribe_topic,
)

_LOGGER = logging.getLogger(__name__)
TIMEOUT = 5
SYS_MODE_MAP = {
    0: ThermostatMode.OFF,
    1: ThermostatMode.AUTO,
    2: ThermostatMode.HEAT,
    3: ThermostatMode.COOL,
}


def _parse_status_flag(status_flag):
    """Parse the status flag."""
    cooling = bit_is_set(status_flag, 0)
    heating = bit_is_set(status_flag, 1)
    energy = bit_is_set(status_flag, 2)
    celcius = bit_is_set(status_flag, 3)
    hold = bit_is_set(status_flag, 4)
    return (cooling, heating, energy, celcius, hold)


PropertyInfo = namedtuple("PropertyInfo", "name group data_field bit set_cmd")


class GetThermostatStatus:
    """Thermostat status manager."""

    def __init__(self, address):
        """Init the GetThermostatStatus class."""
        self._address = Address(address)
        self._get_status_command = ExtendedGetCommand(self._address, cmd2=0x02)
        self._get_set_point_command = ExtendedGetCommand(
            address=self._address, cmd2=0x00, data1=0x00, data2=0x00, data3=0x01
        )
        self._status_response = ExtendedGetResponseHandler(
            self._address, cmd2=0x02, data1=0x01, data2=None, data3=None
        )
        self._set_point_response = ExtendedGetResponseHandler(
            self._address, cmd2=0x00, data1=0x00, data2=0x01, data3=0x01
        )
        self._response_status = asyncio.Queue()
        self._response_set_point = asyncio.Queue()
        self._status_response.subscribe(self._status_received)
        self._set_point_response.subscribe(self._set_point_received)

        self._status_received_topic = (
            f"{self._address.id}.thermostat_status_manager.status_received"
        )
        self._set_point_received_topic = (
            f"{self._address.id}.thermostat_status_manager.set_point_received"
        )
        self._status_listeners = []
        self._set_point_listeners = []

    async def async_status(self, group=None):
        """Read the device status."""
        status1 = await self._status()
        status2 = await self._set_point()
        return multiple_status(status1, status2)

    def subscribe_status(self, listener, force_strong_ref=False):
        """Subscribe to status updates."""
        subscribe_topic(listener=listener, topic_name=self._status_received_topic)
        if force_strong_ref and listener not in self._status_listeners:
            self._status_listeners.append(listener)

    def unsubscribe_status(self, listener):
        """Unsubscribe to status updates."""
        unsubscribe_topic(listener=listener, topic_name=self._status_received_topic)
        if listener in self._status_listeners:
            self._status_listeners.pop(listener)

    def subscribe_set_point(self, listener, force_strong_ref=False):
        """Subscribe to set point updates."""
        subscribe_topic(listener=listener, topic_name=self._set_point_received_topic)
        if force_strong_ref and listener not in self._set_point_listeners:
            self._set_point_listeners.append(listener)

    def unsubscribe_set_point(self, listener):
        """Unsubscribe to set point updates."""
        unsubscribe_topic(listener=listener, topic_name=self._set_point_received_topic)
        if listener in self._set_point_listeners:
            self._set_point_listeners.pop(listener)

    async def _status(self):
        """Send the status command and call set point command."""
        while not self._response_status.empty():
            self._response_status.get_nowait()
        retries = 3
        response_status = ResponseStatus.FAILURE
        while retries:
            response = await self._get_status_command.async_send(crc=True)
            if response == ResponseStatus.SUCCESS:
                try:
                    response_status = await asyncio.wait_for(
                        self._response_status.get(), TIMEOUT
                    )
                    return response_status
                except asyncio.TimeoutError:
                    _LOGGER.debug("Exception _status timed out, retries: %d", retries)
            else:
                response_status = response
            retries -= 1
        return response_status

    async def _set_point(self):
        """Send the set point command."""
        while not self._response_set_point.empty():
            self._response_set_point.get_nowait()
        retries = 3
        response_set_point = ResponseStatus.FAILURE
        while retries:
            response = await self._get_set_point_command.async_send(crc=True)
            if response == ResponseStatus.SUCCESS:
                try:
                    return await asyncio.wait_for(
                        self._response_set_point.get(), TIMEOUT
                    )
                except asyncio.TimeoutError:
                    _LOGGER.debug("Set point response timed out")
            else:
                response_set_point = response
            retries -= 1
        return response_set_point

    def _status_received(self, group, data):
        """Notify the read process that the resonse was received."""
        self._response_status.put_nowait(ResponseStatus.SUCCESS)
        day = data["data2"]
        hour = data["data3"]
        minute = data["data4"]
        second = data["data5"]
        system_mode, fan_mode = calc_thermostat_mode(
            data["data6"], SYS_MODE_MAP, sys_low=False
        )
        cool_set_point = data["data7"]
        humidity = data["data8"]
        temp = calc_thermostat_temp(data["data9"], data["data10"])
        cooling, heating, _, celsius, _ = _parse_status_flag(data["data11"])
        heat_set_point = data["data12"]
        publish_topic(
            self._status_received_topic,
            day=day,
            hour=hour,
            minute=minute,
            second=second,
            system_mode=system_mode,
            fan_mode=fan_mode,
            cool_set_point=cool_set_point,
            humidity=humidity,
            temperature=temp,
            cooling=cooling,
            heating=heating,
            celsius=celsius,
            heat_set_point=heat_set_point,
        )

    def _set_point_received(self, group, data):
        """Handle the Humidity set point response from a device."""
        humidity_high = data["data4"]
        humidity_low = data["data5"]
        firmwire = data["data6"]
        cool_set_point = data["data7"]
        heat_set_point = data["data8"]
        rf_offset = data["data9"]
        publish_topic(
            topic=self._set_point_received_topic,
            humidity_high=humidity_high,
            humidity_low=humidity_low,
            firmwire=firmwire,
            cool_set_point=cool_set_point,
            heat_set_point=heat_set_point,
            rf_offset=rf_offset,
        )
        self._response_set_point.put_nowait(ResponseStatus.SUCCESS)



"""ALDB Read Manager."""
import asyncio
import logging

import async_timeout

from ..address import Address
from ..aldb.aldb_record import ALDBRecord
from ..constants import ALDBStatus, ReadWriteMode, ResponseStatus
from ..data_types.all_link_record_flags import AllLinkRecordFlags
from ..handlers.from_device.receive_aldb_record import ReceiveALDBRecordHandler
from ..handlers.to_device.read_aldb import ReadALDBCommandHandler
from ..managers.peek_poke_manager import get_peek_poke_manager
from ..topics import ALDB_STATUS_CHANGED
from ..utils import subscribe_topic

RETRIES_ALL_MAX = 5
RETRIES_ONE_MAX = 20
TIMER_RECORD = 10
_LOGGER = logging.getLogger(__name__)


def _is_multiple_records(mem_addr, num_recs):
    """Return true if we are searching for multiple records."""
    return (mem_addr == 0x00 and num_recs == 0) or num_recs > 1


class ALDBReadManager:
    """ALDB Read Manager."""

    def __init__(self, address, first_record):
        """Init the ALDBReadManager class."""
        self._address = Address(address)
        self._first_record = first_record
        self._record_queue = asyncio.Queue()
        self._continue = True

        self._read_handler = ReadALDBCommandHandler(self._address)
        self._record_handler = ReceiveALDBRecordHandler(self._address)
        self._record_handler.subscribe(self._receive_record)

        self._peek_bytes_received = asyncio.Queue()
        self._peek_manager = get_peek_poke_manager(self._address)
        self._peek_manager.subscribe_peek(self._receive_peek)
        subscribe_topic(
            self._aldb_status_changed, f"{self._address.id}.{ALDB_STATUS_CHANGED}"
        )

    async def async_read(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0,
        read_write_mode=ReadWriteMode.STANDARD,
    ):
        """Return an iterator of All-Link Database records."""
        _LOGGER.debug(
            "%s: Read memory %04x and %d records",
            str(self._address),
            mem_addr,
            num_recs,
        )
        self._clear_read_queue()
        self._continue = True
        if read_write_mode == ReadWriteMode.PEEK_POKE:
            read_all_method = self._read_all_peek
            read_one_method = self._read_one_peek
            mem_addr = self._first_record if mem_addr == 0 else mem_addr
        else:
            read_all_method = self._read_all
            read_one_method = self._read_one

        if _is_multiple_records(mem_addr, num_recs):
            async for record in read_all_method():
                _LOGGER.debug("Read manager returning: %s", str(record))
                yield record
                await asyncio.sleep(0.05)
        else:
            record = await read_one_method(mem_addr)
            if record is not None:
                _LOGGER.debug("Read manager returning: %s", str(record))
                yield record
                await asyncio.sleep(0.05)

        _LOGGER.debug("Read manager completed")

    async def async_stop(self):
        """Stop the reading process."""
        self._continue = False
        await self._record_queue.put(None)
        await asyncio.sleep(0.01)

    async def _read_one(self, mem_addr):
        """Read one record."""
        retries = RETRIES_ONE_MAX
        while retries and self._continue:
            response = await self._read_handler.async_send(
                mem_addr=mem_addr, num_recs=1
            )
            if response in [
                ResponseStatus.DIRECT_NAK_ALDB,
                ResponseStatus.DIRECT_NAK_INVALID_COMMAND,
                ResponseStatus.DIRECT_NAK_CHECK_SUM,
            ]:
                # When Checksum direct NAK, should we change the checksum method?
                _LOGGER.error(
                    "Device refused the ALDB Read command with error: %s (%r)",
                    response,
                    response,
                )
                return None
            try:
                async with async_timeout.timeout(TIMER_RECORD):
                    record = await self._record_queue.get()
                    if (
                        record is not None
                        and record.mem_addr == mem_addr
                        or mem_addr == 0x0000
                    ):
                        _LOGGER.debug("_read_one returning record: %s", str(record))
                        return record
                    _LOGGER.debug("_read_one not returning record: %s", str(record))
            except asyncio.TimeoutError:
                retries -= 1
            await asyncio.sleep(0.1)
        _LOGGER.debug("_read_one completed")
        return None

    async def _read_one_peek(self, mem_addr):
        """Read one record using peek commands."""
        mem_addr = self._first_record if mem_addr == 0 else mem_addr
        retries = RETRIES_ONE_MAX
        while retries:
            record = bytearray()
            for curr_byte in range(0, 8):
                value = await self._async_peek(mem_addr=mem_addr - curr_byte)
                if value is None:
                    break
                if value == -1:
                    return None
                record.append(value)
            if len(record) == 8:
                flags = AllLinkRecordFlags(record[7])
                aldb_record = ALDBRecord(
                    memory=mem_addr,
                    controller=flags.is_controller,
                    group=record[6],
                    target=Address(bytearray([record[5], record[4], record[3]])),
                    data1=record[2],
                    data2=record[1],
                    data3=record[0],
                    in_use=flags.is_in_use,
                    high_water_mark=flags.is_hwm,
                    bit4=flags.is_bit_4_set,
                    bit5=flags.is_bit_5_set,
                )
                _LOGGER.info(str(aldb_record))
                return aldb_record
            retries -= 1
            _LOGGER.debug("Retrying reading record at 0x%04X", mem_addr)
        _LOGGER.debug("Read ONE return on error - incomplete reads")
        return None

    async def _async_peek(self, mem_addr):
        """Peek one byte."""
        mem_addr = self._first_record if mem_addr == 0 else mem_addr
        _LOGGER.debug("Peeking memory address: 0x%04X", mem_addr)
        retries_byte = 20
        timeout = 3
        while retries_byte:
            while not self._peek_bytes_received.empty():
                await self._peek_bytes_received.get()
            result = await self._peek_manager.async_peek(mem_addr)
            if result == ResponseStatus.SUCCESS:
                try:
                    async with async_timeout.timeout(timeout):
                        peek_addr, value = await self._peek_bytes_received.get()
                        if mem_addr == peek_addr:
                            _LOGGER.debug(
                                "Returning value: %d for memory address: 0x%04X",
                                value,
                                mem_addr,
                            )
                            return value
                except asyncio.TimeoutError:
                    pass
            elif result in [
                ResponseStatus.DIRECT_NAK_ALDB,
                ResponseStatus.DIRECT_NAK_INVALID_COMMAND,
                ResponseStatus.DIRECT_NAK_CHECK_SUM,
            ]:
                # When Checksum direct NAK, should we change the checksum method?
                _LOGGER.error(
                    "Device refused the ALDB Read command with error: %s (%r)",
                    result,
                    result,
                )
                return -1
            retries_byte -= 1
            _LOGGER.debug("Retrying byte at 0x%04X", mem_addr)
        return None

    async def _read_all(self):
        """Read all records."""
        retries = RETRIES_ALL_MAX
        mem_addr = 0
        while retries and self._continue:
            response = await self._read_handler.async_send(
                mem_addr=mem_addr, num_recs=0
            )
            if response in [
                ResponseStatus.DIRECT_NAK_ALDB,
                ResponseStatus.DIRECT_NAK_INVALID_COMMAND,
                ResponseStatus.DIRECT_NAK_CHECK_SUM,
            ]:
                # When Checksum direct NAK, should we change the checksum method?
                _LOGGER.error(
                    "Device refused the ALDB Read command with error: %s (%r)",
                    response,
                    response,
                )
                return
            try:
                while self._continue:
                    async with async_timeout.timeout(TIMER_RECORD):
                        record: ALDBRecord = await self._record_queue.get()
                        if record is None:
                            _LOGGER.debug("_read_all completed")
                            return
                        _LOGGER.debug("_read_all returning record: %s", str(record))
                        if record.is_high_water_mark:
                            mem_addr = 0
                        else:
                            mem_addr = record.mem_addr
                        yield record
                        await asyncio.sleep(0.05)
            except asyncio.TimeoutError:
                retries -= 1

        _LOGGER.debug("_read_all completed")

    async def _read_all_peek(self):
        """Read all ALDB records using peek commands."""
        next_record = self._first_record
        while self._continue:
            record = await self._read_one_peek(next_record)
            if record is None:
                return
            yield record
            await asyncio.sleep(0.05)
            if record.is_high_water_mark:
                return
            await asyncio.sleep(0.06)  # let the ALDB catch up to changes
            next_record = next_record - 8

    async def _receive_record(
        self,
        memory,
        controller,
        group,
        target,
        data1,
        data2,
        data3,
        in_use,
        high_water_mark,
        bit5,
        bit4,
    ):
        """Receive an ALDB record."""
        record = ALDBRecord(
            memory=memory,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        _LOGGER.debug("Putting record in queue: %s", str(record))
        await self._record_queue.put(record)

    async def _receive_peek(self, mem_addr, value):
        """Receive one byte from a peek command."""
        _LOGGER.debug("Putting one byte from peek command")
        await self._peek_bytes_received.put((mem_addr, value))

    def _clear_read_queue(self):
        """Clear the read queue of old records."""

        while not self._record_queue.empty():
            self._record_queue.get_nowait()

    async def _aldb_status_changed(self, status):
        """Listen for status changes and stop the read if needed."""
        if self._continue and status == ALDBStatus.LOADED:
            self._continue = False
            await self._record_queue.put(None)



"""Handle sending a read request for ALDB records."""
import logging

from ...address import Address
from ...constants import AllLinkMode
from ...data_types.all_link_record_flags import AllLinkRecordFlags
from ...topics import ALDB_RECORD_RECEIVED, EXTENDED_READ_WRITE_ALDB
from ...utils import build_topic
from .. import inbound_handler
from ..inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class ReceiveALDBRecordHandler(InboundHandlerBase):
    """Receive an ALDB record direct inbound message."""

    def __init__(self, address: Address):
        """Init the ReceiveALDBRecordHandler class."""
        super().__init__(topic=EXTENDED_READ_WRITE_ALDB, address=address)
        self._subscriber_topic = build_topic(
            prefix="handler",
            address=self._address,
            topic=ALDB_RECORD_RECEIVED,
            message_type="direct",
        )

    @inbound_handler
    def handle_response(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the inbound message."""
        _LOGGER.debug("ALDB Read direct message received")
        if user_data is not None and user_data["d2"] == 0x01:
            memhi = user_data["d3"]
            memlo = user_data["d4"]
            memory = memhi << 8 | memlo

            control_flags = AllLinkRecordFlags(user_data["d6"])

            group = user_data["d7"]

            addrhi = user_data["d8"]
            addrmed = user_data["d9"]
            addrlo = user_data["d10"]
            target = Address(bytearray([addrhi, addrmed, addrlo]))

            data1 = user_data["d11"]
            data2 = user_data["d12"]
            data3 = user_data["d13"]

            self._call_subscribers(
                memory=memory,
                controller=control_flags.link_mode == AllLinkMode.CONTROLLER,
                group=group,
                target=target,
                data1=data1,
                data2=data2,
                data3=data3,
                in_use=control_flags.is_in_use,
                high_water_mark=control_flags.is_hwm,
                bit5=control_flags.is_bit_5_set,
                bit4=control_flags.is_bit_4_set,
            )



"""Extended property read manager."""
import asyncio

import async_timeout

from ..address import Address
from ..constants import ResponseStatus
from ..handlers.from_device.ext_get_response import ExtendedGetResponseHandler
from ..handlers.to_device.extended_get import ExtendedGetCommand
from ..subscriber_base import SubscriberBase

RETRIES = 3
WAIT_RESPONSE = 5


class ExtenededGetManager(SubscriberBase):
    """Extended get command manager."""

    def __init__(
        self,
        address: Address,
        cmd2: int = 0,
        data1_read: int = 0x00,
        data2_read: int = 0x00,
        data3_read: int = None,
        data1_resp: int = 0x00,
        data2_resp: int = 0x01,
        data3_resp: int = None,
    ):
        """Init the ExtenededPropertyReadManager class."""
        self._address = Address(address)
        self._data1_read = data1_read
        self._read_command = ExtendedGetCommand(
            self._address,
            cmd2=cmd2,
            data1=data1_read,
            data2=data2_read,
            data3=data3_read,
        )
        self._response = ExtendedGetResponseHandler(
            self._address,
            cmd2=cmd2,
            data1=data1_resp,
            data2=data2_resp,
            data3=data3_resp,
        )

        response_group = str(cmd2)
        for data in (data1_read, data2_read, data3_read):
            response_group = f"{response_group}.{data if data is not None else '_'}"
        super().__init__(
            f"{self._address.id}.extended_read_manager.{response_group}.response"
        )
        self._response.subscribe(self._response_received)
        self._response_queue = asyncio.Queue()

    async def async_send(self):
        """Send the GET command."""
        retries = RETRIES
        while retries:
            response = await self._read_command.async_send(group=self._data1_read)
            if response == ResponseStatus.SUCCESS:
                response = ResponseStatus.DEVICE_UNRESPONSIVE
                try:
                    async with async_timeout.timeout(WAIT_RESPONSE):
                        return await self._response_queue.get()
                except asyncio.TimeoutError:
                    pass
            retries -= 1
        return response

    async def _response_received(self, group, data):
        """Response received."""
        self._call_subscribers(group=group, data=data)
        await self._response_queue.put(ResponseStatus.SUCCESS)



"""Peek and poke command manager."""

import asyncio
from datetime import datetime, timedelta
import logging

from ..address import Address
from ..constants import ResponseStatus
from ..handlers.to_device.peek import PeekCommand
from ..handlers.to_device.poke import PokeCommand
from ..handlers.to_device.set_msb import SetMsbCommand
from ..subscriber_base import SubscriberBase
from ..topics import PEEK, POKE
from ..utils import publish_topic, subscribe_topic

_instances = {}
_LOGGER = logging.getLogger(__name__)
MSB_WRITE_TIME_DELTA = timedelta(minutes=1)


def get_peek_poke_manager(address: Address):
    """Return a peek/poke manager for a device."""
    address = Address(address)
    instance = _instances.get(address)
    if instance:
        return instance
    instance = PeekPokeManager(address)
    _instances[address] = instance
    return instance


class PeekPokeManager(SubscriberBase):
    """Peek and poke command manager."""

    def __init__(self, address):
        """Init the PeekPokeManager class."""
        self._address = Address(address)
        super().__init__(f"{self._address.id}.manager.peek_poke")
        self._peek_cmd = PeekCommand(self._address)
        self._poke_cmd = PokeCommand(self._address)
        self._set_msb_cmd = SetMsbCommand(self._address)
        self._peek_value_queue = asyncio.Queue()
        self._poke_value_queue = asyncio.Queue()
        self._peek_cmd.subscribe(self._receive_peek_value)
        self._peek_topic = f"{self._address.id}.manager.{PEEK}"
        self._poke_topic = f"{self._address.id}.manager.{POKE}"
        self._last_msb = None
        self._time_last_msb = datetime.min

    async def async_peek(self, mem_addr: int, extended: bool = False):
        """Peek a value at a memory address."""
        result = await self._async_peek(mem_addr=mem_addr)
        if result == ResponseStatus.SUCCESS:
            value = await self._peek_value_queue.get()
            publish_topic(topic=self._peek_topic, mem_addr=mem_addr, value=value)
            await asyncio.sleep(0.05)
        return result

    async def async_poke(self, mem_addr: int, value: int):
        """Poke a value at a memory address."""
        if 0 > value > 255:
            raise ValueError("Poke value can only be one byte.")
        result = await self._async_peek(mem_addr)
        if result == ResponseStatus.SUCCESS:
            orig_value = await self._peek_value_queue.get()
            if orig_value == value:
                publish_topic(topic=self._poke_topic, mem_addr=mem_addr, value=value)
                await asyncio.sleep(0.05)
                return ResponseStatus.SUCCESS
            result = await self._poke_cmd.async_send(value=value)
            if result == ResponseStatus.SUCCESS:
                publish_topic(topic=self._poke_topic, mem_addr=mem_addr, value=value)
                await asyncio.sleep(0.05)
        return result

    def subscribe_peek(self, listener: callable, force_strong_ref: bool = False):
        """Subscribe to the PEEK results."""
        if force_strong_ref and listener not in self._subscribers:
            _LOGGER.debug("Adding subscriber to persistant list")
            self._subscribers.append(listener)
        subscribe_topic(listener=listener, topic_name=self._peek_topic)

    def subscribe_poke(self, listener: callable, force_strong_ref: bool = False):
        """Subscribe to the POKE results."""
        if force_strong_ref and listener not in self._subscribers:
            _LOGGER.debug("Adding subscriber to persistant list")
            self._subscribers.append(listener)
        subscribe_topic(listener=listener, topic_name=self._poke_topic)

    async def _async_peek(self, mem_addr: int):
        """Execute the peek command."""
        msb_set = False
        mem_hi = mem_addr >> 8
        mem_lo = mem_addr & 0xFF
        while not self._peek_value_queue.empty():
            await self._peek_value_queue.get()

        if (
            mem_hi == self._last_msb
            and self._time_last_msb > datetime.now() - MSB_WRITE_TIME_DELTA
        ):
            msb_set = True
        else:
            result = await self._set_msb_cmd.async_send(high_byte=mem_hi)
            if result == ResponseStatus.SUCCESS:
                self._last_msb = mem_hi
                self._time_last_msb = datetime.now()
                msb_set = True
                await asyncio.sleep(0.1)
        if msb_set:
            result = await self._peek_cmd.async_send(lsb=mem_lo)
        return result

    async def _receive_peek_value(self, value):
        """Place the last peek value in the queue."""
        _LOGGER.debug("Received peek value: %d (0x%02x)", value, value)
        await self._peek_value_queue.put(value)
        await asyncio.sleep(0.05)

    async def _receive_poke_value(self, value):
        """Place the last poke value in the queue."""
        _LOGGER.debug("Received poke value: %d (0x%02x)", value, value)
        await self._poke_value_queue.put(value)
        await asyncio.sleep(0.05)



"""Insteon device address class."""
import logging

from .constants import HC_LOOKUP
from .utils import (
    byte_to_housecode,
    byte_to_unitcode,
    housecode_to_byte,
    unitcode_to_byte,
)

_LOGGER = logging.getLogger(__name__)


def _normalize(addr):
    """Take any format of address and turn it into a byte."""

    def to_housecode_unitcode(hc_uc_byte):
        """Convert a byte value to a housecode and unitcode."""
        hc_uc = int.from_bytes(hc_uc_byte, byteorder="big")
        housecode = hc_uc >> 4
        unitcode = hc_uc & 0x0F
        return housecode, unitcode

    if isinstance(addr, X10Address):
        return addr.housecode_byte, addr.unitcode_byte

    if isinstance(addr, (bytearray, bytes)):
        return to_housecode_unitcode(bytes(addr))

    if isinstance(addr, int):
        return to_housecode_unitcode(bytes([addr]))

    if isinstance(addr, str):
        addr_clean = addr.replace(".", "").lower()
        if len(addr_clean) == 6:
            housecode = housecode_to_byte(addr_clean[3])
            unitcode = unitcode_to_byte(int(addr_clean[4:]))
            return housecode, unitcode
    raise ValueError(f"Improper X10 address: {addr}")


def create(housecode: str, unitcode: int):
    """Create an X10 device address."""
    if housecode.lower() in HC_LOOKUP:
        byte_housecode = housecode_to_byte(housecode)
    else:
        if isinstance(housecode, str):
            str_error = f"X10 house code invalid: {housecode}"
        else:
            str_error = "X10 house code is not a string"
            raise ValueError(str_error)

    # 20, 21 and 22 for All Units Off, All Lights On and All Lights Off
    # 'fake' units
    if unitcode in range(1, 17):
        byte_unitcode = unitcode_to_byte(unitcode)
    elif unitcode == 0:
        byte_unitcode = 0x00
    else:
        if isinstance(unitcode, int):
            str_error = f"X10 unit code error: {unitcode}"
        else:
            str_error = "X10 unit code is not an integer 1 - 16"
            raise ValueError(str_error)

    addr = X10Address((byte_housecode << 4) + byte_unitcode)
    return addr


class X10Address:
    """Datatype definition for an X10 device address."""

    def __init__(self, housecode_unitcode: (bytes, bytearray, str, int)):
        """Create an X10 device address."""
        housecode, unitcode = _normalize(housecode_unitcode)
        self._housecode_byte = housecode
        self._unitcode_byte = unitcode
        if not self._check_housecode_unitcode():
            raise ValueError("Invalid housecode or unitcode byte")

    def __repr__(self):
        """Return the string representation of an X10 device address."""
        housecode = byte_to_housecode(self._housecode_byte).lower()
        unitcode = byte_to_unitcode(self._unitcode_byte)
        return f"x10{housecode}{unitcode:02d}"

    def __str__(self):
        """Return the string representation of an X10 device address."""
        str_rep = f"X10.{self.housecode}.{self.unitcode:02d}"
        return str(str_rep)

    def __bytes__(self):
        """Return the byte representation of an X10 address."""
        int_value = (self._housecode_byte << 4) + self._unitcode_byte
        return bytes(bytearray([int_value]))

    def __eq__(self, other):
        """Test if two X10 addresses are equal."""
        if isinstance(other, X10Address):
            return bytes(self) == bytes(other)
        return False

    def __getitem__(self, byte):
        """Return the housecode or unitcode bytes."""
        if byte == 0:
            return self.housecode_byte
        if byte == 1:
            return self.unitcode_byte
        err = f"Item index must be 0 or 1: {byte}"
        raise ValueError(err)

    def __hash__(self):
        """Return a hash of the address."""
        return hash(str)

    def _check_housecode_unitcode(self):
        """Check to confirm housecode and unitcode are valid."""
        housecode = byte_to_housecode(self._housecode_byte)
        unitcode = byte_to_unitcode(self._unitcode_byte)
        if housecode and unitcode:
            return True
        return False

    @property
    def id(self):
        """Return the ID of the X10 address."""
        return repr(self)

    @property
    def housecode_byte(self):
        """Emit the X10 house code byte value."""
        return self._housecode_byte

    @property
    def unitcode_byte(self):
        """Emit the X10 unit code byte value."""
        return self._unitcode_byte

    @property
    def housecode(self):
        """Emit the X10 house code."""
        return byte_to_housecode(self._housecode_byte).upper()

    @property
    def unitcode(self):
        """Emit the X10 unit code."""
        return byte_to_unitcode(self._unitcode_byte)



"""Representaton of an extended property of a device."""

from ..constants import PropertyType
from .device_flag import DeviceFlagBase


class ExtendedProperty(DeviceFlagBase):
    """Representation of an extended property of a device."""

    def __init__(
        self,
        address,
        name,
        value_type: type,
        is_reversed=False,
        is_read_only=False,
        prop_type=PropertyType.STANDARD,
    ):
        """Init the ExtendedProperty class."""
        super().__init__(
            address, "property", name, value_type, is_reversed, is_read_only, prop_type
        )



"""ALDB Records data class."""

from ..address import Address
from ..constants import AllLinkMode


class ALDBRecord:
    """Represents an ALDB record."""

    def __init__(
        self,
        memory: int,
        controller: bool,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
        in_use: bool = True,
        high_water_mark: bool = False,
        bit5: bool = False,
        bit4: bool = False,
    ):
        """Initialze the ALDBRecord class."""
        self._memory_location = memory
        self._target = Address(target) if target is not None else target
        self._group = group
        self._data1 = data1
        self._data2 = data2
        self._data3 = data3
        self._controller = controller
        self._in_use = in_use
        self._high_water_mark = high_water_mark
        self._bit5 = bit5
        self._bit4 = bit4

    def __str__(self):
        """Return the string representation of an ALDB record."""
        if self._controller:
            link_mode = "C"
        else:
            link_mode = "R"
        rec = {
            "memory": f"0x{self._memory_location:04x}",
            "inuse": self._in_use,
            "link_mode": link_mode,
            "bit5": str(self._bit5),
            "bit4": str(self._bit4),
            "highwater": str(self._high_water_mark),
            "group": f"0x{int(self.group):02x}",
            "target": str(self.target),
            "data1": f"0x{int(self.data1):02x}",
            "data2": f"0x{int(self.data2):02x}",
            "data3": f"0x{int(self.data3):02x}",
        }
        return str(rec)

    def __repr__(self):
        """Return a representation of the record."""
        rec = {
            "memory": f"0x{self._memory_location:04x}",
            "control_flags": f"0x{int(self.control_flags):02x}",
            "group": f"0x{int(self.group):02x}",
            "target": str(self.target),
            "data1": f"0x{int(self.data1):02x}",
            "data2": f"0x{int(self.data2):02x}",
            "data3": f"0x{int(self.data3):02x}",
        }
        return str(rec)

    def __eq__(self, other: object) -> bool:
        """Test if two records are equal.

        Two records are equal if the following match:
            Mode (responder or controller)
            Group
            Target
            Data 3 if the record is a responder
        """
        if not isinstance(other, ALDBRecord):
            return False

        return (
            (
                other.mode is None
                or self.mode is None
                or other.is_controller == self.is_controller
            )
            and (other.group is None or self.group is None or other.group == self.group)
            and (
                other.target is None
                or self.target is None
                or other.target == self.target
            )
            and (
                self.mode == AllLinkMode.CONTROLLER
                or (
                    (other.group == 0 and self.group == 0)
                    or other.data3 is None
                    or self.data3 is None
                    or self.data3 == other.data3
                )
            )
        )

    def __iter__(self):
        """Return a dictionary object of the ALDB Record."""

        return_dict = {
            "d1": 0x00,
            "d2": 0x00,
            "d3": self.memhi,
            "d4": self.memlo,
            "d5": 0x00,
            "d6": self.control_flags,
            "d7": self.group,
            "d8": self.target.high,
            "d9": self.target.middle,
            "d10": self.target.low,
            "d11": self.data1,
            "d12": self.data2,
            "d13": self.data3,
            "d14": 0x00,
        }
        for key, value in return_dict.items():
            yield (key, value)

    @property
    def mem_addr(self):
        """Return the memory address of the database record."""
        return self._memory_location

    @mem_addr.setter
    def mem_addr(self, value):
        """Set the memory address of the record."""
        try:
            mem = int(value)
        except ValueError as ex:
            raise ValueError("Memory address must be an integer.") from ex
        self._memory_location = mem

    @property
    def memhi(self):
        """Return the memory address MSB."""
        return self._memory_location >> 8

    @property
    def memlo(self):
        """Return the memory address LSB."""
        return self._memory_location & 0xFF

    @property
    def target(self):
        """Return the address of the device the record points to."""
        return self._target

    @property
    def group(self):
        """Return the group the record responds to."""
        return self._group

    @property
    def mode(self) -> AllLinkMode:
        """Return the All-Link mode."""
        if self._controller is None:
            return None
        return AllLinkMode.CONTROLLER if self._controller else AllLinkMode.RESPONDER

    @property
    def data1(self):
        """Return the data1 field of the ALDB record."""
        return self._data1

    @property
    def data2(self):
        """Return the data2 field of the ALDB record."""
        return self._data2

    @property
    def data3(self):
        """Return the data3 field of the ALDB record."""
        return self._data3

    @property
    def is_controller(self):
        """Return if the record is a controller record."""
        return self._controller

    @property
    def is_responder(self):
        """Return if the record is a responder."""
        return not self._controller

    @property
    def is_in_use(self):
        """Return if the record is in use."""
        return self._in_use

    @property
    def is_high_water_mark(self):
        """Return if this is the high water mark record."""
        return self._high_water_mark

    @property
    def is_bit5_set(self):
        """Return if control flag bit 5 is set."""
        return self._bit5

    @property
    def is_bit4_set(self):
        """Return if control flag bit 4 is set."""
        return self._bit4

    @property
    def control_flags(self):
        """Return the control flag byte."""
        return (
            int(self._in_use) << 7
            | int(self._controller) << 6
            | int(self._bit5) << 5
            | int(self._bit4) << 4
            | int(not self._high_water_mark) << 1
        )

    def copy(self):
        """Create a new record as a copy of the current record."""
        return ALDBRecord(
            memory=self._memory_location,
            controller=self._controller,
            group=self._group,
            target=self._target,
            data1=self._data1,
            data2=self._data2,
            data3=self._data3,
            in_use=self._in_use,
            bit5=self._bit5,
            bit4=self._bit4,
        )

    def is_exact_match(self, rec, test_in_use=True):
        """Return True if the input record exactly matches this record."""
        try:
            if (
                self == rec
                and (self.is_in_use == rec.is_in_use or not test_in_use)
                and self.data1 == rec.data1
                and self.data2 == rec.data2
            ):
                return True
        except AttributeError:
            return False
        return False


def new_aldb_record_from_existing(rec: ALDBRecord, **kwargs):
    """Create a new ALDB record from an existing record."""

    def get_element_or_default(key: str, default: any, **kwargs):
        """Return the value of the key or the default.

        Need to use this since the kwarg.get(key, default) can return
        `None` even if the default is not None if kwargs[key] is None.
        """
        value = kwargs.get(key)
        if value is not None:
            return value
        return default

    mem_addr = int(get_element_or_default("mem_addr", rec.mem_addr, **kwargs))
    in_use = bool(get_element_or_default("in_use", rec.is_in_use, **kwargs))
    controller = bool(get_element_or_default("controller", rec.is_controller, **kwargs))
    hwm = bool(
        get_element_or_default("high_water_mark", rec.is_high_water_mark, **kwargs)
    )
    target = Address(get_element_or_default("target", rec.target, **kwargs))
    group = int(get_element_or_default("group", rec.group, **kwargs))
    data1 = int(get_element_or_default("data1", rec.data1, **kwargs))
    data2 = int(get_element_or_default("data2", rec.data2, **kwargs))
    data3 = int(get_element_or_default("data3", rec.data3, **kwargs))
    bit5_set = bool(get_element_or_default("bit5", rec.is_bit5_set, **kwargs))
    bit4_set = bool(get_element_or_default("bit4", rec.is_bit4_set, **kwargs))

    return ALDBRecord(
        memory=mem_addr,
        controller=controller,
        group=group,
        target=target,
        high_water_mark=hwm,
        data1=data1,
        data2=data2,
        data3=data3,
        in_use=in_use,
        bit5=bit5_set,
        bit4=bit4_set,
    )



"""IOLink momentary delay property."""
from . import get_usable_value
from .derived_property import DerivedProperty


class MomentaryDelayProperty(DerivedProperty):
    """IOLink momentary delay property."""

    def __init__(
        self,
        address,
        name,
        delay_prop,
        prescaler_prop,
    ):
        """Init the OnOffDelayProperty class."""
        super().__init__(address, name, float, False, False)

        self._delay_prop = delay_prop
        self._prescaler_prop = prescaler_prop

    @property
    def value(self):
        """Return the momentary delay value of the underlying proprties."""
        prescaler = max(self._prescaler_prop.value, 1)
        return self._delay_prop.value * prescaler / 10

    @property
    def new_value(self):
        """Return the new_value property."""
        if not (self._prescaler_prop.is_dirty or self._delay_prop.is_dirty):
            return None
        prescaler = get_usable_value(self._prescaler_prop)
        delay = get_usable_value(self._delay_prop)
        prescaler = max(prescaler, 1)
        return delay * prescaler / 10

    @new_value.setter
    def new_value(self, value):
        """Set the new_value property."""
        seconds = self._value_type(value)

        if not 0 < seconds < 6502:  # 255^2 / 10  or about 1 hour 48 min
            raise ValueError

        if seconds == self.value:
            self._delay_prop.new_value = None
            self._prescaler_prop.new_value = None
            return

        delay = seconds * 10
        prescaler = 1
        if delay > 255:
            prescaler = int(round(delay / 255 + 0.5, 0))
            delay = int(round(delay / prescaler, 0))
        self._prescaler_prop.new_value = prescaler
        self._delay_prop.new_value = delay

    @property
    def is_dirty(self):
        """Return the change status of the underlying properties."""
        return self._delay_prop.is_dirty or self._prescaler_prop.is_dirty

    @property
    def is_loaded(self):
        """Return the load status of the underlying properties."""
        return self._delay_prop.is_loaded and self._prescaler_prop.is_loaded



"""Constants used to ensure topic consistantancy."""

DEVICE_LIST_CHANGED = "device_list_changed"

STANDARD_RECEIVED = "standard_message_received"
EXTENDED_RECEIVED = "extended_message_received"
X10_RECEIVED = "x10_received"
ALL_LINKING_COMPLETED = "all_linking_completed"
BUTTON_EVENT_REPORT = "button_event_report"
USER_RESET_DETECTED = "user_reset_detected"
ALL_LINK_CLEANUP_FAILURE_REPORT = "all_link_cleanup_failure_report"
ALL_LINK_RECORD_RESPONSE = "all_link_record_response"
ALL_LINK_CLEANUP_STATUS_REPORT = "all_link_cleanup_status_report"
GET_IM_INFO = "get_im_info"
SEND_ALL_LINK_COMMAND = "send_all_link_command"
SEND_STANDARD = "send_standard"
SEND_EXTENDED = "send_extended"
X10_SEND = "x10_send"
START_ALL_LINKING = "start_all_linking"
CANCEL_ALL_LINKING = "cancel_all_linking"
SET_HOST_DEVICE_CATEGORY = "set_host_device_category"
RESET_IM = "reset_im"
SET_ACK_MESSAGE_BYTE = "set_ack_message_byte"
GET_FIRST_ALL_LINK_RECORD = "get_first_all_link_record"
GET_NEXT_ALL_LINK_RECORD = "get_next_all_link_record"
SET_IM_CONFIGURATION = "set_im_configuration"
GET_ALL_LINK_RECORD_FOR_SENDER = "get_all_link_record_for_sender"
LED_ON = "led_on"
LED_OFF = "led_off"
MANAGE_ALL_LINK_RECORD = "manage_all_link_record"
SET_NAK_MESSAGE_BYTE = "set_nak_message_byte"
SET_ACK_MESSAGE_TWO_BYTES = "set_ack_message_two_bytes"
RF_SLEEP = "rf_sleep"
GET_IM_CONFIGURATION = "get_im_configuration"
READ_EEPROM = "read_eeprom"
READ_EEPROM_RESPONSE = "read_eeprom_response"
WRITE_EEPROM = "write_eeprom"

# Command Topics
ASSIGN_TO_ALL_LINK_GROUP = "assign_to_all_link_group"
DELETE_FROM_ALL_LINK_GROUP = "delete_from_all_link_group"
PRODUCT_DATA_REQUEST = "product_data_request"
FX_USERNAME = "fx_username"
DEVICE_TEXT_STRING_REQUEST = "device_text_string_request"
SET_DEVICE_TEXT_STRING = "set_device_text_string"
SET_ALL_LINK_COMMAND_ALIAS = "set_all_link_command_alias"
SET_ALL_LINK = "set_all_link"
ENTER_LINKING_MODE = "enter_linking_mode"
ENTER_UNLINKING_MODE = "enter_unlinking_mode"
GET_INSTEON_ENGINE_VERSION = "get_insteon_engine_version"
PING = "ping"
ID_REQUEST = "id_request"
ID_REQUEST_RESPONSE = "id_request_response"
ON = "on"
ON_FAST = "on_fast"
OFF = "off"
OFF_FAST = "off_fast"
BRIGHTEN_ONE_STEP = "brighten_one_step"
DIM_ONE_STEP = "dim_one_step"
START_MANUAL_CHANGE_DOWN = "start_manual_change_down"
START_MANUAL_CHANGE_UP = "start_manual_change_up"
STOP_MANUAL_CHANGE = "stop_manual_change"
STATUS_REQUEST = "status_request"
GET_OPERATING_FLAGS = "get_operating_flags"
SET_OPERATING_FLAGS = "set_operating_flags"
INSTANT_CHANGE = "instant_change"
MANUALLY_TURNED_OFF = "manually_turned_off"
MANUALLY_TURNED_ON = "manually_turned_on"
REMOTE_SET_BUTTON_TAP1_TAP = "remote_set_button_tap1_tap"
REMOTE_SET_BUTTON_TAP2_TAP = "remote_set_button_tap2_tap"
SET_STATUS = "set_status"
SET_ADDRESS_MSB = "set_address_msb"
POKE = "poke"
PEEK = "peek"
PEEK_INTERNAL = "peek_internal"
POKE_INTERNAL = "poke_internal"
ON_AT_RAMP_RATE = "on_at_ramp_rate"
ON_AT_RAMP_RATE_INBOUND = "on_at_ramp_rate_inbound"
EXTENDED_GET_SET = "extended_get_set"
EXTENDED_GET_RESPONSE = "extended_get_response"
OFF_AT_RAMP_RATE = "off_at_ramp_rate"
OFF_AT_RAMP_RATE_INBOUND = "off_at_ramp_rate_inbound"
EXTENDED_READ_WRITE_ALDB = "extended_read_write_aldb"
EXTENDED_TRIGGER_ALL_LINK = "extended_trigger_all_link"
ALDB_RECORD_RECEIVED = "aldb_record_received"
BEEP = "beep"
GROUP_ON = "group_on"
GROUP_OFF = "group_off"
FACTORY_RESET = "factory_reset"
NIGHT_MODE_ON = "night_mode_on"
NIGHT_MODE_OFF = "night_mode_off"
SPRINKLER_VALVE_ON = "sprinkler_valve_on"
SPRINKLER_VALVE_OFF = "sprinkler_valve_off"
SPRINKLER_PROGRAM_ON = "sprinkler_program_on"
SPRINKLER_PROGRAM_OFF = "sprinkler_program_off"
SPRINKLER_CONTROL = "sprinkler_control"
SPRINKLER_GET_PROGRAM_REQUEST = "sprinkler_get_program_request"
IO_OUTPUT_ON = "io_output_on"
IO_OUTPUT_OFF = "io_output_off"
IO_ALARM_DATA_REQUEST = "io_alarm_data_request"
IO_WRITE_OUTPUT_PORT = "io_write_output_port"
IO_READ_INPUT_PORT = "io_read_input_port"
IO_GET_SENSOR_VALUE = "io_get_sensor_value"
IO_SET_SENSOR_1_NOMINAL_VALUE = "io_set_sensor_1_nominal_value"
IO_GET_SENSOR_ALARM_DELTA = "io_get_sensor_alarm_delta"
FAN_STATUS_REPORT = "fan_status_report"
IO_WRITE_CONFIGURATION_PORT = "io_write_configuration_port"
IO_READ_CONFIGURATION_PORT = "io_read_configuration_port"
IO_MODULE_CONTROL = "io_module_control"
POOL_DEVICE_ON = "pool_device_on"
POOL_DEVICE_OFF = "pool_device_off"
POOL_TEMPERATURE_UP = "pool_temperature_up"
POOL_TEMPERATURE_DOWN = "pool_temperature_down"
POOL_CONTROL = "pool_control"
POOL_SET_DEVICE_TEMPERATURE = "pool_set_device_temperature"
POOL_SET_DEVICE_HYSTERESIS = "pool_set_device_hysteresis"
DOOR_CONTROL = "door_control"
DOOR_STATUS_REPORT = "door_status_report"
WINDOW_COVERING_CONTROL = "window_covering_control"
WINDOW_COVERING_POSITION = "window_covering_position"
THERMOSTAT_TEMPERATURE_UP = "thermostat_temperature_up"
THERMOSTAT_TEMPERATURE_DOWN = "thermostat_temperature_down"
THERMOSTAT_GET_ZONE_INFORMATION = "thermostat_get_zone_information"
THERMOSTAT_CONTROL = "thermostat_control"
THERMOSTAT_SET_COOL_SETPOINT = "thermostat_set_cool_setpoint"
THERMOSTAT_SET_HEAT_SETPOINT = "thermostat_set_heat_setpoint"
THERMOSTAT_EXTENDED_STATUS = "thermostat_extended_status"
THERMOSTAT_TEMPERATURE_STATUS = "thermostat_temperature_status"
THERMOSTAT_HUMIDITY_STATUS = "thermostat_humidity_status"
THERMOSTAT_MODE_STATUS = "thermostat_mode_status"
THERMOSTAT_COOL_SET_POINT_STATUS = "thermostat_cool_set_point_status"
THERMOSTAT_HEAT_SET_POINT_STATUS = "thermostat_heat_set_point_status"
LEAK_DETECTOR_ANNOUNCE = "leak_detector_announce"
SET_SPRINKLER_PROGRAM = "set_sprinkler_program"
SPRINKLER_GET_PROGRAM_RESPONSE = "sprinkler_get_program_response"
IO_SET_SENSOR_NOMINAL_VALUE = "io_set_sensor_nominal_value"
IO_ALARM_DATA_RESPONSE = "io_alarm_data_response"

ALDB_LINK_CHANGED = "aldb.link_changed"

ENGINE_VERSION = "engine_version"
ALDB_STATUS_CHANGED = "aldb_status_changed"
EXTENDED_PROPERTIES_CHANGED = "extended_properties_changed"


GET_QUEUE_SIZE = "get_queue_size"
SEND_QUEUE_SIZE = "send_queue_size"
ADD_DEFAULT_LINKS = "add_default_links"



"""Manages links between devices to identify device state of responders."""
import asyncio
from datetime import datetime, timedelta
from typing import Dict, Tuple, Union

import voluptuous as vol

from .. import pub
from ..address import Address
from ..aldb.aldb_record import ALDBRecord
from ..constants import AllLinkMode, DeviceAction, MessageFlagType, ResponseStatus
from ..topics import ALDB_LINK_CHANGED
from ..utils import subscribe_topic, unsubscribe_topic

TIMEOUT_DUPLICATE = timedelta(seconds=10)
ControllerAddress = Address
ResponderAddress = Address
Group = int

DeviceLinkSchema = vol.Schema(
    [
        {
            vol.Required("address"): str,
            vol.Required("data1"): int,
            vol.Required("data2"): int,
            vol.Required("data3"): int,
        }
    ]
)


class LinkInfo:
    """Device link data class."""

    def __init__(self, data1, data2, data3, has_controller, has_responder):
        """Init the LinkInfo class."""
        self.data1 = data1
        self.data2 = data2
        self.data3 = data3
        self.has_controller = has_controller
        self.has_responder = has_responder


EMPTY_ADDRESS = Address("000000")


def _add_controller_group_responder_tree(links, controller, group, responder):
    """Add a controller, group, responder set."""
    if not links.get(controller):
        links[controller] = {}
    if not links[controller].get(group):
        links[controller][group] = {}
    if not links[controller][group].get(responder):
        links[controller][group][responder] = []
    return links


def _add_controller_link(links, controller, record):
    """Add a controller link."""
    return _add_controller_group_responder_tree(
        links, controller, record.group, record.target
    )


def _add_responder_link(links, devices, responder, record):
    """Add a responder link."""
    links = _add_controller_group_responder_tree(
        links, record.target, record.group, responder
    )
    responder_links = links[record.target][record.group][responder]
    controller_device = devices[record.target]
    has_controller = False
    if controller_device:
        for _ in controller_device.aldb.find(
            target=responder,
            is_controller=True,
            group=record.group,
        ):
            has_controller = True
    link = LinkInfo(record.data1, record.data2, record.data3, has_controller, True)
    responder_links.append(link)
    return links


def _topic_to_addr_group(topic):
    elements = topic.name.split(".")
    try:
        address = Address(elements[0])
        group = int(elements[1])
        command = elements[2]
        msg_type = getattr(MessageFlagType, elements[-1].upper())
    except [KeyError, TypeError]:
        return None, None, None, None
    return address, group, command, msg_type


class DeviceLinkManager:
    """Manages links between devices to identify state of responders.

    Listens for broadcast messages from a device and identifies the responders of the device.
    Once the responders are identified, the state of the responders is determined.
    """

    def __init__(self, devices):
        """Init the DeviceLinkManager class."""
        self._devices = devices
        self._work_dir: Union[str, None] = None
        self._devices.subscribe(self._device_added_or_removed)
        # The _last_command property holds the last command to change a device group
        # If the last command is the same as the current command, do nothing.
        # Otherwise we send a status request
        self._last_command: Dict[Address, Dict[int, Tuple[str, datetime]]] = {}

    @property
    def links(
        self,
    ) -> Dict[ControllerAddress, Dict[Group, Dict[ResponderAddress, LinkInfo]]]:
        """Return a list of device links."""
        return self._get_links()

    def get_responders(
        self, controller: Address, group: int
    ) -> Dict[ResponderAddress, LinkInfo]:
        """Return the responders to a controller/group combination."""
        responders = self._get_links(controller, group)
        return responders.get(controller, {}).get(group, {})

    def _get_links(self, controller=None, group=None):
        """Return a list of links.

        If controller is not None -> Only return responders for that controller
        If group is not None -> Only return responders for that group
        """
        modem_address = self._devices.modem.address if self._devices.modem else None
        links = {}
        for addr in self._devices:
            if addr == modem_address:
                continue
            device = self._devices[addr]
            for mem_addr in device.aldb:
                rec: ALDBRecord = device.aldb[mem_addr]
                if (
                    rec.target == modem_address
                    or rec.target == EMPTY_ADDRESS
                    or not rec.is_in_use
                ):
                    continue
                if rec.mode == AllLinkMode.CONTROLLER:
                    if not controller or device.address == controller:
                        links = _add_controller_link(links, device.address, rec)
                    continue
                if (not controller or rec.target == controller) and (
                    not group or rec.group != group
                ):
                    links = _add_responder_link(
                        links, self._devices, device.address, rec
                    )
        return links

    def _link_changed(self, record: ALDBRecord, sender: Address, deleted: bool) -> None:
        """Add a record to the controller/responder list."""
        if record.is_controller:
            controller = Address(sender)
            responder = Address(record.target)
        else:
            controller = Address(record.target)
            responder = Address(sender)

        if deleted:
            self._remove_link(record=record, controller=controller)
        else:
            self._add_link(record=record, controller=controller, responder=responder)

    def _add_link(self, record, controller, responder):
        """Add a link to the controller/responder list."""
        if self._is_standard_modem_link(controller, responder, record.group):
            return
        # Listen for the controller group topic and check known responders.
        subscribe_topic(self._async_check_responders, f"{controller.id}.{record.group}")

    def _remove_link(self, record, controller):
        """Remove a controller or responder link from the controller/responder list."""
        responders = self.get_responders(controller, record.group)
        if not responders:
            unsubscribe_topic(
                self._async_check_responders, f"{controller.id}.{record.group}"
            )

    async def _device_added_or_removed(self, address: Address, action: DeviceAction):
        """Track device list changes."""
        await asyncio.sleep(0.1)
        if action == DeviceAction.REMOVED:
            unsubscribe_topic(self._link_changed, f"{address.id}.{ALDB_LINK_CHANGED}")

        elif action == DeviceAction.ADDED:
            device = self._devices[address]
            if not device:
                await asyncio.sleep(0.1)
                device = self._devices[address]
                if not device:
                    subscribe_topic(
                        self._link_changed, f"{address.id}.{ALDB_LINK_CHANGED}"
                    )
                    return
            device.aldb.subscribe_record_changed(self._link_changed)
            for _, record in device.aldb.items():
                if record.is_in_use:
                    self._link_changed(record=record, sender=address, deleted=False)

    async def _async_check_responders(self, topic=pub.AUTO_TOPIC, **kwargs) -> None:
        controller, group, command, msg_type = _topic_to_addr_group(topic)

        if msg_type not in (
            MessageFlagType.ALL_LINK_BROADCAST,
            MessageFlagType.ALL_LINK_CLEANUP,
        ):
            return
        if group == 0:
            return
        if self._is_duplicate_message(controller, group, command):
            return
        responder_data = self.get_responders(controller, group)
        for addr, data_list in responder_data.items():
            device = self._devices[addr]
            if device:
                # If the device is a category 1 or 2 device we can pre-load the device state with the
                # ALDB record data1 field value. We will then check the actual status later.
                for data in data_list:
                    if device.cat in [0x01, 0x02] and data.data3 is not None:
                        device.groups[data.data3].value = data.data1
                if not device.is_battery:
                    response = ResponseStatus.UNSENT
                    retries = 5
                    while retries and response != ResponseStatus.SUCCESS:
                        response = await device.async_status()
                        retries -= 1

    def _is_standard_modem_link(self, controller, responder, group):
        """Test if a link is a standard modem link."""
        if self._devices.modem and responder == self._devices.modem.address:
            return True
        if self._devices.modem and controller == self._devices.modem.address:
            device = self._devices[responder]
            if device:
                groups = list(device.groups)
            groups.append(0)
            if group in groups:
                return True
        return False

    def _is_duplicate_message(self, controller: Address, group: int, command: str):
        """Test if this is the duplicate message to notify of a change.

        This method is intended to reduce the number of status request changes to responders.
        """
        last_command = self._last_command.get(controller, {}).get(group)
        if not last_command:
            self._save_last_command(controller, group, command)
            return False
        tdelta = datetime.now() - last_command[1]
        if command == last_command[0] and tdelta < TIMEOUT_DUPLICATE:
            return True
        self._save_last_command(controller, group, command)
        return False

    def _save_last_command(self, controller: Address, group: int, command: str):
        """Save the date and time of the last command from a controller."""
        controller_commands = self._last_command.get(controller, {})
        controller_commands[group] = (command, datetime.now())
        self._last_command[controller] = controller_commands



"""Device state definitions."""
DIMMABLE_LIGHT = "dimmable_light"
DIMMABLE_FAN = "dimmable_fan"
DIMMABLE_OUTLET = "dimmable_outlet"
ON_OFF_SWITCH = "on_off_switch"
RELAY = "RELAY"
ON_OFF_OUTLET_TOP = "on_off_outlet_top"
ON_OFF_OUTLET_BOTTOM = "on_off_outlet_bottom"
LOW_BATTERY = "low_battery"
OPEN_CLOSE_SENSOR = "open_close_sensor"
LIGHT_SENSOR = "light_sensor"
LOAD_SENSOR = "load_sensor"
LEAK_SENSOR_DRY = "leak_sensor_dry"
LEAK_SENSOR_WET = "leak_sensor_wet"
DOOR_SENSOR = "door_sensor"
MOTION_SENSOR = "motion_sensor"
SMOKE_SENSOR = "smoke_sensor"
CO_SENSOR = "co_sensor"
TEST_SENSOR = "test_sensor"
NEW_SENSOR = "new_sensor"
HEARTBEAT = "heartbeat"
SENSOR_MALFUNCTION = "sensor_malfunction"
COVER = "cover"
COOL_SET_POINT = "cool_set_point"
HEAT_SET_POINT = "heat_set_point"
HUMIDITY_HIGH = "humidity_high"
HUMIDITY_LOW = "humidity_low"
SYSTEM_MODE = "system_mode"
FAN_MODE = "fan_mode"
TEMPERATURE = "temperature"
HUMIDITY = "humidity"
COOLING = "cooling"
HEATING = "heating"
HUMIDIFYING = "humidifying"
DEHUMIDIFYING = "dehumidifying"

ON_OFF_SWITCH_MAIN = f"{ON_OFF_SWITCH}_main"
ON_OFF_SWITCH_A = f"{ON_OFF_SWITCH}_a"
ON_OFF_SWITCH_B = f"{ON_OFF_SWITCH}_b"
ON_OFF_SWITCH_C = f"{ON_OFF_SWITCH}_c"
ON_OFF_SWITCH_D = f"{ON_OFF_SWITCH}_d"
ON_OFF_SWITCH_E = f"{ON_OFF_SWITCH}_e"
ON_OFF_SWITCH_F = f"{ON_OFF_SWITCH}_f"
ON_OFF_SWITCH_G = f"{ON_OFF_SWITCH}_g"
ON_OFF_SWITCH_H = f"{ON_OFF_SWITCH}_h"

DIMMABLE_LIGHT_MAIN = f"{DIMMABLE_LIGHT}_main"



"""Fan On Level state."""

from ..address import Address
from ..constants import FanSpeed, FanSpeedRange
from .group_base import GroupBase


class FanOnLevel(GroupBase):
    """On / Off state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: FanSpeed = None,
        status_type: int = 0,
    ):
        """Init the FanOnLevel class."""
        super().__init__(
            name, address, group, default, value_type=FanSpeed, status_type=status_type
        )
        self._is_dimmable = True

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        if on_level in FanSpeedRange.OFF:
            fan_speed = FanSpeed.OFF
        elif on_level in FanSpeedRange.LOW:
            fan_speed = FanSpeed.LOW
        elif on_level in FanSpeedRange.MEDIUM:
            fan_speed = FanSpeed.MEDIUM
        else:
            fan_speed = FanSpeed.HIGH
        self.value = fan_speed



"""Start All-Linking."""
from . import ack_handler, nak_handler
from ..constants import AllLinkMode
from ..topics import START_ALL_LINKING
from .outbound_base import OutboundHandlerBase


# pylint: disable=arguments-differ
class StartAllLinkingCommandHandler(OutboundHandlerBase):
    """Start All-Linking Command."""

    def __init__(self):
        """Init the StartAllLinkingCommandHandler class."""
        super().__init__(topic=START_ALL_LINKING)

    async def async_send(self, link_mode: AllLinkMode, group: int = 0):
        """Send the Start All-Linking Command."""
        return await super().async_send(link_mode=link_mode, group=group)

    @ack_handler
    async def async_handle_ack(self, link_mode, group):
        """Handle the ACK response."""
        return await self._async_handle_ack()

    @nak_handler
    async def async_handle_nak(self, link_mode, group):
        """Handle the NAK response."""
        await self._async_handle_nak()



"""Dimmable Lighting Control Devices (CATEGORY 0x01)."""

import asyncio
from collections.abc import Iterable
from functools import partial
from typing import Dict, Union

from ..config import (
    CLEANUP_REPORT_ON,
    CRC_ERROR_COUNT,
    DATABASE_DELTA,
    KEY_BEEP_ON,
    LED_BLINK_ON_ERROR_OFF,
    LED_BLINK_ON_ERROR_ON,
    LED_BLINK_ON_TX_ON,
    LED_DIMMING,
    LED_OFF,
    LOAD_BUTTON,
    LOAD_BUTTON_NUMBER,
    LOAD_SENSE_ON,
    NIGHT_MODE_ON,
    NO_CACHE,
    NON_TOGGLE_MASK,
    NON_TOGGLE_ON_OFF_MASK,
    OFF_MASK,
    ON_LEVEL,
    ON_MASK,
    POWERLINE_DISABLE_ON,
    PROGRAM_LOCK_ON,
    RADIO_BUTTON_GROUPS,
    RAMP_RATE,
    RAMP_RATE_IN_SEC,
    RED_LED_OFF,
    RELAY_AT_FULL_ON,
    RELAY_MODE_OFF,
    RESUME_DIM_ON,
    RF_DISABLE_ON,
    SIGNAL_TO_NOISE_FAILURE_COUNT,
    SKIP_SOME_HOPS,
    TOGGLE_BUTTON,
    TRIGGER_GROUP_MASK,
    X10_HOUSE,
    X10_OFF,
    X10_UNIT,
    YAKETY_YAK,
)
from ..config.extended_property import ExtendedProperty
from ..config.load_button import LoadButtonProperty
from ..config.radio_button import RadioButtonGroupsProperty
from ..config.ramp_rate import RampRateProperty
from ..config.toggle_button import ToggleButtonProperty
from ..constants import FanSpeed, PropertyType, ResponseStatus, ToggleMode
from ..events import OFF_EVENT, OFF_FAST_EVENT, ON_EVENT, ON_FAST_EVENT
from ..groups import (
    DIMMABLE_FAN,
    DIMMABLE_LIGHT,
    DIMMABLE_LIGHT_MAIN,
    DIMMABLE_OUTLET,
    ON_OFF_SWITCH_A,
    ON_OFF_SWITCH_B,
    ON_OFF_SWITCH_C,
    ON_OFF_SWITCH_D,
    ON_OFF_SWITCH_E,
    ON_OFF_SWITCH_F,
    ON_OFF_SWITCH_G,
    ON_OFF_SWITCH_H,
)
from ..groups.on_level import OnLevel
from ..groups.on_off import OnOff
from ..handlers.from_device.manual_change import ManualChangeInbound
from ..handlers.to_device.group_off import GroupOffCommand
from ..handlers.to_device.set_leds import SetLedsCommandHandler
from ..handlers.to_device.trigger_scene_off import TriggerSceneOffCommandHandler
from ..handlers.to_device.trigger_scene_on import TriggerSceneOnCommandHandler
from ..utils import bit_is_set, multiple_status, set_bit, set_fan_speed
from .device_commands import (
    MANUAL_CHANGE,
    OFF_COMMAND,
    OFF_FAST_COMMAND,
    ON_COMMAND,
    ON_FAST_COMMAND,
    SET_LEDS_COMMAND,
    STATUS_COMMAND,
)
from .i3_base import I3VariableResponderBase
from .variable_controller_base import ON_LEVEL_MANAGER
from .variable_responder_base import VariableResponderBase


class DimmableLightingControl(VariableResponderBase):
    """Dimmable Lighting Control Device."""

    def _register_handlers_and_managers(self):
        """Register command handlers and managers."""
        super()._register_handlers_and_managers()
        for group, group_prop in self._groups.items():
            if isinstance(group_prop, OnLevel):
                self._handlers[group][MANUAL_CHANGE] = ManualChangeInbound(
                    self._address, group
                )

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe methods to handlers and managers."""
        super()._subscribe_to_handelers_and_managers()
        self._managers[STATUS_COMMAND].remove_status_type(0)
        self._managers[STATUS_COMMAND].add_status_type(2, self._handle_status)
        for group, group_prop in self._groups.items():
            if isinstance(group_prop, OnLevel):
                self._handlers[group][MANUAL_CHANGE].subscribe(
                    self._async_on_manual_change
                )

    async def _async_on_manual_change(self, group=None):
        """Respond to a manual change of the device."""
        await self.async_status()


class DimmableLightingControl_LampLinc(DimmableLightingControl):
    """LampLinc based dimmable lights."""

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_OFF, 0, 4, 8, 9)
        self._add_operating_flag(LOAD_SENSE_ON, 0, 5, 0x0A, 0x0B)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None, prop_type=PropertyType.ADVANCED)
        self._add_property(X10_UNIT, 6, None, prop_type=PropertyType.ADVANCED)
        self._add_property(RAMP_RATE, 7, 5, prop_type=PropertyType.ADVANCED)
        self._add_property(ON_LEVEL, 8, 6)

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )


class DimmableLightingControl_SwitchLincBase(DimmableLightingControl):
    """SwichLinc based dimmable lights."""

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_OFF, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15)

        self._add_property(X10_HOUSE, 5, None, prop_type=PropertyType.ADVANCED)
        self._add_property(X10_UNIT, 6, None, prop_type=PropertyType.ADVANCED)
        self._add_property(RAMP_RATE, 7, 5, prop_type=PropertyType.ADVANCED)
        self._add_property(ON_LEVEL, 8, 6)

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )


class DimmableLightingControl_SwitchLinc01(DimmableLightingControl_SwitchLincBase):
    """SwichLinc based dimmable lights.

    Uses command 0x2E 0x00 0x00 0x03 for LED dimming.
    """

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()
        if (self._subcat == 0x04 and self._firmware >= 0x30) or (
            self._subcat == 0x19 and self._firmware >= 0x41
        ):
            self._add_property(LED_DIMMING, 9, 7)
        else:
            self._add_property(LED_DIMMING, 3, 3)


class DimmableLightingControl_SwitchLinc02(DimmableLightingControl_SwitchLincBase):
    """SwichLinc based dimmable lights.

    Uses command 0x2E 0x00 0x00 0x07 for LED dimming.
    """

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()
        self._add_property(LED_DIMMING, 9, 7)


class DimmableLightingControl_ToggleLinc(DimmableLightingControl_SwitchLinc01):
    """SwichLinc based dimmable lights."""


class DimmableLightingControl_InLineLinc01(DimmableLightingControl_SwitchLinc01):
    """InLineLinc based dimmable lights.

    Uses command 0x2E 0x00 0x00 0x03 for LED dimming.
    """


class DimmableLightingControl_InLineLinc02(DimmableLightingControl_SwitchLinc02):
    """InLineLinc based dimmable lights.

    Uses command 0x2E 0x00 0x00 0x07 for LED dimming.
    """


class DimmableLightingControl_OutletLinc(DimmableLightingControl):
    """OutletLinc based dimmable lights."""

    def __init__(self, address, cat, subcat, firmware=0, description="", model=""):
        """Init the DimmableLightingControl_OutletLinc class."""
        buttons = {1: (DIMMABLE_OUTLET, 0)}
        super().__init__(
            address,
            cat,
            subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_OFF, 0, 4, 8, 9)

        self._add_property(X10_HOUSE, 5, None, prop_type=PropertyType.ADVANCED)
        self._add_property(X10_UNIT, 6, None, prop_type=PropertyType.ADVANCED)


class DimmableLightingControl_DinRail(DimmableLightingControl):
    """DINRail based dimmable lights."""

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_OFF, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)

        self._add_property(X10_HOUSE, 5, None, prop_type=PropertyType.ADVANCED)
        self._add_property(X10_UNIT, 6, None, prop_type=PropertyType.ADVANCED)
        self._add_property(RAMP_RATE, 7, 5, prop_type=PropertyType.ADVANCED)
        self._add_property(ON_LEVEL, 8, 6)
        self._add_property(LED_DIMMING, 9, 7)

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )


class DimmableLightingControl_FanLinc(DimmableLightingControl):
    """FanLinc model 2475F Dimmable Lighting Control.

    Device Class 0x01 subcat 0x2e

    """

    def __init__(self, address, cat, subcat, firmware=0, description="", model=""):
        """Init the DimmableLightingControl_FanLinc class."""
        buttons = {1: (DIMMABLE_LIGHT, 0), 2: (DIMMABLE_FAN, 3)}
        super().__init__(
            address,
            cat,
            subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )

    def fan_on(self, on_level: FanSpeed = FanSpeed.HIGH, fast=False):
        """Turn on the fan.

        Parameters:
            on_level: Default FanSpeed.HIGH (full on). Set the fan on level.
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        group = 2
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        self._handlers[group][command].send(on_level=int(set_fan_speed(on_level)))

    async def async_fan_on(self, on_level: FanSpeed = FanSpeed.HIGH, fast=False):
        """Turn on the device.

        Parameters:
            on_level: Default FanSpeed.HIGH (full on). Set the fan on level.
            fast: Default False. If True, bypass fan ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        group = 2
        command = ON_FAST_COMMAND if fast else ON_COMMAND
        return await self._handlers[group][command].async_send(
            on_level=int(set_fan_speed(on_level))
        )

    def fan_off(self, fast=False):
        """Turn off the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        """
        group = 2
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        self._handlers[group][command].send()

    async def async_fan_off(self, fast=False):
        """Turn off the device.

        Parameters:
            fast: Default False. If True, bypass device ramp rate otherwise
            turn on at the ramp rate.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        group = 2
        command = OFF_FAST_COMMAND if fast else OFF_COMMAND
        return await self._handlers[group][command].async_send()

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._managers[STATUS_COMMAND].add_status_type(3, self._handle_fan_status)

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()

        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_OFF, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(RF_DISABLE_ON, 0, 6, 0x0C, 0x0D)
        self._add_operating_flag(POWERLINE_DISABLE_ON, 0, 7, 0x0E, 0x0F)

        self._add_operating_flag(DATABASE_DELTA, 1, None, None, None)
        self._add_operating_flag(CRC_ERROR_COUNT, 2, None, None, None)
        self._add_operating_flag(SIGNAL_TO_NOISE_FAILURE_COUNT, 3, None, None, None)

        self._add_operating_flag(X10_OFF, 5, 1, 0x12, 0x13)
        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15)
        self._add_operating_flag(CLEANUP_REPORT_ON, 5, 3, 0x16, 0x17)

        self._add_property(X10_HOUSE, 5, None, prop_type=PropertyType.ADVANCED)
        self._add_property(X10_UNIT, 6, None, prop_type=PropertyType.ADVANCED)
        self._add_property(RAMP_RATE, 7, 5, prop_type=PropertyType.ADVANCED)
        self._add_property(ON_LEVEL, 8, 6)

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )

    def _handle_fan_status(self, db_version, status):
        self._groups[2].set_value(status)


class DimmableLightingControl_KeypadLinc(DimmableLightingControl):
    """KeypadLinc base class."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the GeneralController_MiniRemoteBase class."""
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )
        self._on_off_lock = asyncio.Lock()

    async def async_on(self, on_level: int = 0xFF, group: int = 0, fast: bool = False):
        """Turn on the button LED."""
        if group in [0, 1]:
            result = await super().async_on(on_level=on_level, group=group, fast=fast)
        else:
            async with self._on_off_lock:
                kwargs = self._calc_led_status(led=group, is_on=True)
                result = await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)
        if result == ResponseStatus.SUCCESS:
            event = ON_FAST_EVENT if fast else ON_EVENT
            self._update_led_state(group=group, value=on_level, event=event)
        elif result == ResponseStatus.DIRECT_NAK_PRE_NAK:
            await self.async_status()
        return result

    async def async_off(self, group: int = 0, fast: bool = False):
        """Turn off the button LED."""
        if group in [0, 1]:
            result = await super().async_off(group=group, fast=fast)
        else:
            async with self._on_off_lock:
                kwargs = self._calc_led_status(led=group, is_on=False)
                result = await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)
        if result == ResponseStatus.SUCCESS:
            event = OFF_FAST_EVENT if fast else OFF_EVENT
            self._update_led_state(group=group, value=0, event=event)
        elif result == ResponseStatus.DIRECT_NAK_PRE_NAK:
            await self.async_status()
        return result

    async def async_trigger_scene_on(
        self, group: int, on_level: int = 255, fast: bool = False
    ):
        """Trigger an All-Link group on."""
        cmd = TriggerSceneOnCommandHandler(self._address, group)
        return await cmd.async_send(on_level, fast)

    async def async_trigger_scene_off(self, group: int, fast: bool = False):
        """Trigger an All-Link group off."""
        cmd = TriggerSceneOffCommandHandler(self._address, group)
        return await cmd.async_send(fast)

    def set_radio_buttons(self, buttons: Iterable):
        """Set a group of buttons to act as radio buttons.

        This takes in a iterable set of buttons (eg. (3,4,5,6)) to act as radio buttons where
        no two buttons are on at the same time.
        """
        if len(buttons) < 2:
            raise IndexError("At least two buttons required.")

        for button in buttons:
            if button not in self._buttons.keys():
                raise ValueError(f"Button {button} not in button list.")
            button_str = f"_{button}" if button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            if not on_mask.is_loaded or not off_mask.is_loaded:
                on_mask.set_value(0)
                off_mask.set_value(0)
            on_mask_new_value = 0
            off_mask_new_value = 0
            for bit in range(0, 8):
                if bit + 1 in buttons:
                    on_mask_value = bit != button - 1
                    off_mask_value = bit != button - 1
                else:
                    on_mask_value = bit_is_set(on_mask.value, bit)
                    off_mask_value = bit_is_set(off_mask.value, bit)
                on_mask_new_value = set_bit(on_mask_new_value, bit, on_mask_value)
                off_mask_new_value = set_bit(off_mask_new_value, bit, off_mask_value)
            on_mask.new_value = on_mask_new_value
            off_mask.new_value = off_mask_new_value

    def clear_radio_buttons(self, buttons: Iterable):
        """Clear the radio button behavior of the button.

        This takes in a single button number or a collection of buttons.
        For any button received, the radio button behavior of that button and
        any button that is grouped with that button will be cleared.

        Example:
            - Buttons C and D are currently radio buttons
            - Call `clear_radio_buttons(3)` which represents the C button.
            - Because C and D are currently grouped as radio buttons,
              both C and D will have their on and off masks changed to clear the
              link to the other button.

        """
        other_buttons = [
            button for button in self._buttons if button not in buttons and button != 1
        ]
        addl_buttons = []
        for other_button in other_buttons:
            button_str = f"_{other_button}" if other_button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            if not on_mask.is_loaded or not off_mask.is_loaded:
                on_mask.set_value(0)
                off_mask.set_value(0)
            for button in buttons:
                bit = button - 1
                on_set = (
                    bit_is_set(on_mask.new_value, bit)
                    if on_mask.is_dirty
                    else bit_is_set(on_mask.value, bit)
                )
                off_set = (
                    bit_is_set(off_mask.new_value, bit)
                    if off_mask.is_dirty
                    else bit_is_set(off_mask.value, bit)
                )
                if on_set or off_set and other_button not in addl_buttons:
                    addl_buttons.append(other_button)
                    continue

        for button in buttons:
            button_str = f"_{button}" if button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            on_mask.new_value = 0
            off_mask.new_value = 0

        for addl_button in addl_buttons:
            button_str = f"_{addl_button}" if addl_button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            if not on_mask.is_loaded or not off_mask.is_loaded:
                on_mask.set_value(0)
                off_mask.set_value(0)
            for button in buttons:
                if on_mask.is_dirty:
                    on_mask.new_value = set_bit(on_mask.new_value, button - 1, False)
                else:
                    on_mask.new_value = set_bit(on_mask.value, button - 1, False)
                if off_mask.is_dirty:
                    off_mask.new_value = set_bit(off_mask.new_value, button - 1, False)
                else:
                    off_mask.new_value = set_bit(off_mask.value, button - 1, False)

    def set_toggle_mode(self, button: int, toggle_mode: ToggleMode):
        """Set the toggle mode of a button.

        Usage:
            button: Integer of the button number
            toggle_mode: Integer of the toggle mode
                0: Toggle
                1: Non-Toggle ON only
                2: Non-Toggle OFF only
        """
        if button not in self._buttons.keys():
            raise ValueError(f"Button {button} not in button list.")
        try:
            toggle_mode = ToggleMode(toggle_mode)
        except ValueError as err:
            raise ValueError(
                f"Toggle mode {toggle_mode} invalid. Valid modes are [0, 1, 2]"
            ) from err

        toggle_mask = self.properties[NON_TOGGLE_MASK]
        on_off_mask = self.properties[NON_TOGGLE_ON_OFF_MASK]
        if not toggle_mask.is_loaded or not on_off_mask.is_loaded:
            toggle_mask.set_value(0)
            on_off_mask.set_value(0)

        if toggle_mask.new_value is None:
            toggle_mask_test = toggle_mask.value
        else:
            toggle_mask_test = toggle_mask.new_value

        if on_off_mask.new_value is None:
            on_off_mask_test = on_off_mask.value
        else:
            on_off_mask_test = on_off_mask.new_value

        if toggle_mode == ToggleMode.TOGGLE:
            toggle_mask.new_value = set_bit(toggle_mask_test, button - 1, False)
            on_off_mask.new_value = set_bit(on_off_mask_test, button - 1, False)
        elif toggle_mode == ToggleMode.ON_ONLY:
            toggle_mask.new_value = set_bit(toggle_mask_test, button - 1, True)
            on_off_mask.new_value = set_bit(on_off_mask_test, button - 1, True)
        else:
            toggle_mask.new_value = set_bit(toggle_mask_test, button - 1, True)
            on_off_mask.new_value = set_bit(on_off_mask_test, button - 1, False)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[SET_LEDS_COMMAND] = SetLedsCommandHandler(address=self.address)

    def _register_groups(self):
        for button in self._buttons:
            name = self._buttons[button][0]
            status_type = self._buttons[button][1]
            if button == 1:
                self._groups[button] = OnLevel(
                    name=name,
                    address=self._address,
                    group=button,
                    status_type=status_type,
                )
            else:
                self._groups[button] = OnOff(
                    name=name,
                    address=self._address,
                    group=button,
                    status_type=status_type,
                )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._managers[STATUS_COMMAND].add_status_type(1, self._led_status)
        for group in self._buttons:
            if group != 1:
                led_method = partial(self._async_led_follow_check, group=group)
                self._managers[group][ON_LEVEL_MANAGER].subscribe(led_method)

    def _async_led_follow_check(self, group, on_level):
        """Check the other LEDs to confirm if they follow the effected LED."""
        for button in self._buttons:
            if button == group:
                continue
            button_str = f"_{button}" if button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            if not on_mask.is_loaded or not off_mask.is_loaded:
                continue
            follow = bit_is_set(on_mask.value, group)
            set_off = bit_is_set(off_mask.value, group)
            if follow:
                if set_off:
                    self._groups[button].value = 0
                else:
                    self._groups[button].value = on_level

    def _calc_led_status(self, led, is_on):
        leds = {}
        for curr_led in range(1, 9):
            var = f"group{curr_led}"
            curr_group = self._groups.get(curr_led)
            curr_val = bool(curr_group.value) if curr_group else False
            leds[var] = is_on if curr_led == led else curr_val
        return leds

    def _update_led_state(self, group, value, event):
        """Check if the LED is toggle or not and set value."""
        if not self._groups.get(group):
            return
        if self._properties[NON_TOGGLE_MASK].value is not None:
            non_toogle = bit_is_set(self._properties[NON_TOGGLE_MASK].value, group)
        else:
            non_toogle = False
        if non_toogle:
            self._groups[group].value = 0
        else:
            self._groups[group].value = value
        self._events[group][event].trigger(value)

    def _led_status(self, db_version, status):
        """Set the on level of the LED from a status command."""
        for bit in range(2, 9):
            state = self._groups.get(bit)
            if state:
                state.value = bit_is_set(status, bit - 1)

    def _register_op_flags_and_props(self):
        """Register operating flags."""
        super()._register_op_flags_and_props()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_OFF, 0, 4, 8, 9)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(RF_DISABLE_ON, 0, 6, 0x0C, 0x0D)
        self._add_operating_flag(POWERLINE_DISABLE_ON, 0, 7, 0x0E, 0x0F)
        self._add_operating_flag(LED_BLINK_ON_ERROR_OFF, 5, 2, 0x14, 0x15)

        self._add_property(LED_DIMMING, 9, 7, 1)
        self._add_property(NON_TOGGLE_MASK, 0x0A, 0x08, prop_type=PropertyType.ADVANCED)
        self._add_property(
            NON_TOGGLE_ON_OFF_MASK, 0x0D, 0x0B, prop_type=PropertyType.ADVANCED
        )
        self._add_property(
            TRIGGER_GROUP_MASK, 0x0E, 0x0C, prop_type=PropertyType.ADVANCED
        )
        for button in self._buttons:
            button_str = f"_{button}" if button != 1 else ""
            self._add_property(
                f"{ON_MASK}{button_str}", 3, 2, button, prop_type=PropertyType.ADVANCED
            )
            self._add_property(
                f"{OFF_MASK}{button_str}", 4, 3, button, prop_type=PropertyType.ADVANCED
            )
            self._add_property(
                f"{X10_HOUSE}{button_str}",
                5,
                None,
                button,
                prop_type=PropertyType.ADVANCED,
            )
            self._add_property(
                f"{X10_UNIT}{button_str}",
                6,
                None,
                button,
                prop_type=PropertyType.ADVANCED,
            )
            self._add_property(
                f"{RAMP_RATE}{button_str}",
                7,
                5,
                button,
                prop_type=PropertyType.ADVANCED,
            )
            self._add_property(
                f"{ON_LEVEL}{button_str}", 8, 6, button, prop_type=PropertyType.ADVANCED
            )

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )
        self._config[RADIO_BUTTON_GROUPS] = RadioButtonGroupsProperty(
            self, RADIO_BUTTON_GROUPS
        )
        for group in self._groups:
            if group == 1:
                continue
            button = self._buttons[group]
            name = f"{TOGGLE_BUTTON}_{button[-1]}"
            self._config[name] = ToggleButtonProperty(
                self._address,
                name,
                group,
                self.properties[NON_TOGGLE_MASK],
                self.properties[NON_TOGGLE_ON_OFF_MASK],
            )


class DimmableLightingControl_KeypadLinc_6(DimmableLightingControl_KeypadLinc):
    """KeypadLinc 6 button dimmer."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the DimmableLightingControl_KeypadLinc_6 class."""
        buttons = {
            1: (DIMMABLE_LIGHT_MAIN, 0),
            3: (ON_OFF_SWITCH_A, 1),
            4: (ON_OFF_SWITCH_B, 1),
            5: (ON_OFF_SWITCH_C, 1),
            6: (ON_OFF_SWITCH_D, 1),
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )


class DimmableLightingControl_KeypadLinc_8(DimmableLightingControl_KeypadLinc):
    """KeypadLinc 8 button dimmer."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the DimmableLightingControl_KeypadLinc_6 class."""
        buttons = {
            1: (DIMMABLE_LIGHT_MAIN, 0),
            2: (ON_OFF_SWITCH_B, 1),
            3: (ON_OFF_SWITCH_C, 1),
            4: (ON_OFF_SWITCH_D, 1),
            5: (ON_OFF_SWITCH_E, 1),
            6: (ON_OFF_SWITCH_F, 1),
            7: (ON_OFF_SWITCH_G, 1),
            8: (ON_OFF_SWITCH_H, 1),
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )


class DimmableLightingControl_Dial(I3VariableResponderBase):
    """Dimmable i3 Dial."""

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe methods to handlers and managers."""
        super()._subscribe_to_handelers_and_managers()
        self._managers[STATUS_COMMAND].remove_status_type(0)
        self._managers[STATUS_COMMAND].add_status_type(2, self._handle_status)
        for group, group_prop in self._groups.items():
            if isinstance(group_prop, OnLevel):
                self._handlers[group][MANUAL_CHANGE].subscribe(
                    self._async_on_manual_change
                )

    async def _async_on_manual_change(self, group=None):
        """Respond to a manual change of the device."""
        await self.async_status()

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )


class DimmableLightingControl_I3_KeypadLinc_4(I3VariableResponderBase):
    """I3 dimmable 4 button KeypadLink device."""

    _is_kpl = True
    _op_flags_data_4 = {
        0: YAKETY_YAK,
        1: RED_LED_OFF,
        2: SKIP_SOME_HOPS,
        3: RELAY_AT_FULL_ON,
        4: f"{NIGHT_MODE_ON}_kpl",
        5: f"{NO_CACHE}_kpl",
    }

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the DimmableLightingControl_I3_KeypadLink_4 class."""
        buttons = {
            1: (ON_OFF_SWITCH_A, 1),
            2: (ON_OFF_SWITCH_B, 1),
            3: (ON_OFF_SWITCH_C, 1),
            4: (ON_OFF_SWITCH_D, 1),
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )
        self._on_off_lock = asyncio.Lock()

    @property
    def load_button(self):
        """Return the button that controls the load."""
        load_button = self._properties[LOAD_BUTTON_NUMBER].value
        load_button = 1 if load_button in [0, None] else load_button
        return load_button

    async def async_on(self, on_level: int = 0xFF, group: int = 0, fast: bool = False):
        """Turn on the button LED."""
        group = self.load_button if not group else group
        led_kwargs = {}
        event = ON_FAST_EVENT if fast else ON_EVENT
        result_load = ResponseStatus.SUCCESS
        result_led = ResponseStatus.SUCCESS
        if group != self.load_button:
            load_new_state = self._get_led_follow_state(
                button1=group, button2=self.load_button, new_state=on_level
            )
            if load_new_state != bool(self._groups[self.load_button].value):
                if load_new_state:
                    result_load = await super().async_on(fast=fast)
                else:
                    result_load = await super().async_off(fast=fast)
        if group == self.load_button:
            result_load = await super().async_on(
                on_level=on_level, group=group, fast=fast
            )
        if group != self.load_button or self._has_follow_leds(group):
            async with self._on_off_lock:
                led_kwargs = self._calc_set_leds_command_kwargs(
                    led=group, is_on=bool(on_level)
                )
                result_led = await self._handlers[SET_LEDS_COMMAND].async_send(
                    **led_kwargs
                )

        result = multiple_status(result_load, result_led)
        if result == ResponseStatus.SUCCESS:
            await self._async_update_led_state(
                group=group, value=on_level, event=event, led_kwargs=led_kwargs
            )
        else:
            await self.async_status()
        return result

    async def async_off(self, group: int = 0, fast: bool = False):
        """Turn off the button LED."""
        group = self.load_button if not group else group
        led_kwargs = {}
        event = OFF_FAST_EVENT if fast else OFF_EVENT
        result_load = ResponseStatus.SUCCESS
        result_led = ResponseStatus.SUCCESS
        if group == self.load_button:
            result_load = await super().async_off(group=group, fast=fast)
        if group != self.load_button or self._has_follow_leds(group):
            async with self._on_off_lock:
                led_kwargs = self._calc_set_leds_command_kwargs(led=group, is_on=False)
                result_led = await self._handlers[SET_LEDS_COMMAND].async_send(
                    **led_kwargs
                )

        result = multiple_status(result_load, result_led)
        if result == ResponseStatus.SUCCESS:
            await self._async_update_led_state(
                group=group, value=0, event=event, led_kwargs=led_kwargs
            )
        else:
            await self.async_status()
        return result

    async def async_trigger_scene_on(
        self, group: int, on_level: int = 255, fast: bool = False
    ):
        """Trigger an All-Link group on."""
        cmd = TriggerSceneOnCommandHandler(self._address, group)
        return await cmd.async_send(on_level, fast)

    async def async_trigger_scene_off(self, group: int, fast: bool = False):
        """Trigger an All-Link group off."""
        cmd = TriggerSceneOffCommandHandler(self._address, group)
        return await cmd.async_send(fast)

    async def async_group_off(self, group: int) -> ResponseStatus:
        """Trigger a group off."""
        cmd = GroupOffCommand(address=self._address)
        return await cmd.async_send(group=group)

    def set_radio_buttons(self, buttons: Iterable):
        """Set a group of buttons to act as radio buttons.

        This takes in a iterable set of buttons (eg. (3,4,5,6)) to act as radio buttons where
        no two buttons are on at the same time.
        """
        if len(buttons) < 2:
            raise IndexError("At least two buttons required.")

        for button in buttons:
            if button not in self._buttons.keys():
                raise ValueError(f"Button {button} not in button list.")
            button_str = f"_{button}" if button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            if not on_mask.is_loaded or not off_mask.is_loaded:
                on_mask.set_value(0)
                off_mask.set_value(0)
            on_mask_new_value = 0
            off_mask_new_value = 0
            for bit in range(0, 8):
                if bit + 1 in buttons:
                    on_mask_value = bit != button - 1
                    off_mask_value = bit != button - 1
                else:
                    on_mask_value = bit_is_set(on_mask.value, bit)
                    off_mask_value = bit_is_set(off_mask.value, bit)
                on_mask_new_value = set_bit(on_mask_new_value, bit, on_mask_value)
                off_mask_new_value = set_bit(off_mask_new_value, bit, off_mask_value)
            on_mask.new_value = on_mask_new_value
            off_mask.new_value = off_mask_new_value

    def clear_radio_buttons(self, buttons: Iterable):
        """Clear the radio button behavior of the button.

        This takes in a single button number or a collection of buttons.
        For any button received, the radio button behavior of that button and
        any button that is grouped with that button will be cleared.

        Example:
            - Buttons C and D are currently radio buttons
            - Call `clear_radio_buttons(3)` which represents the C button.
            - Because C and D are currently grouped as radio buttons,
              both C and D will have their on and off masks changed to clear the
              link to the other button.

        """
        other_buttons = [
            button for button in self._buttons if button not in buttons and button != 1
        ]
        addl_buttons = []
        for other_button in other_buttons:
            button_str = f"_{other_button}" if other_button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            if not on_mask.is_loaded or not off_mask.is_loaded:
                on_mask.set_value(0)
                off_mask.set_value(0)
            for button in buttons:
                bit = button - 1
                on_set = (
                    bit_is_set(on_mask.new_value, bit)
                    if on_mask.is_dirty
                    else bit_is_set(on_mask.value, bit)
                )
                off_set = (
                    bit_is_set(off_mask.new_value, bit)
                    if off_mask.is_dirty
                    else bit_is_set(off_mask.value, bit)
                )
                if on_set or off_set and other_button not in addl_buttons:
                    addl_buttons.append(other_button)
                    continue

        for button in buttons:
            button_str = f"_{button}" if button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            on_mask.new_value = 0
            off_mask.new_value = 0

        for addl_button in addl_buttons:
            button_str = f"_{addl_button}" if addl_button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            if not on_mask.is_loaded or not off_mask.is_loaded:
                on_mask.set_value(0)
                off_mask.set_value(0)
            for button in buttons:
                if on_mask.is_dirty:
                    on_mask.new_value = set_bit(on_mask.new_value, button - 1, False)
                else:
                    on_mask.new_value = set_bit(on_mask.value, button - 1, False)
                if off_mask.is_dirty:
                    off_mask.new_value = set_bit(off_mask.new_value, button - 1, False)
                else:
                    off_mask.new_value = set_bit(off_mask.value, button - 1, False)

    def set_toggle_mode(self, button: int, toggle_mode: ToggleMode):
        """Set the toggle mode of a button.

        Usage:
            button: Integer of the button number
            toggle_mode: Integer of the toggle mode
                0: Toggle
                1: Non-Toggle ON only
                2: Non-Toggle OFF only
        """
        if button not in self._buttons.keys():
            raise ValueError(f"Button {button} not in button list.")
        try:
            toggle_mode = ToggleMode(toggle_mode)
        except ValueError as err:
            raise ValueError(
                f"Toggle mode {toggle_mode} invalid. Valid modes are [0, 1, 2]"
            ) from err

        toggle_mask = self.properties[NON_TOGGLE_MASK]
        on_off_mask = self.properties[NON_TOGGLE_ON_OFF_MASK]
        if not toggle_mask.is_loaded or not on_off_mask.is_loaded:
            toggle_mask.set_value(0)
            on_off_mask.set_value(0)

        if toggle_mask.new_value is None:
            toggle_mask_test = toggle_mask.value
        else:
            toggle_mask_test = toggle_mask.new_value

        if on_off_mask.new_value is None:
            on_off_mask_test = on_off_mask.value
        else:
            on_off_mask_test = on_off_mask.new_value

        if toggle_mode == ToggleMode.TOGGLE:
            toggle_mask.new_value = set_bit(toggle_mask_test, button - 1, False)
            on_off_mask.new_value = set_bit(on_off_mask_test, button - 1, False)
        elif toggle_mode == ToggleMode.ON_ONLY:
            toggle_mask.new_value = set_bit(toggle_mask_test, button - 1, True)
            on_off_mask.new_value = set_bit(on_off_mask_test, button - 1, True)
        else:
            toggle_mask.new_value = set_bit(toggle_mask_test, button - 1, True)
            on_off_mask.new_value = set_bit(on_off_mask_test, button - 1, False)

    def _register_groups(self):
        """Register the groups."""
        super()._register_groups()
        for group, state in self._groups.items():
            if group != 1:
                state.is_dimmable = False

    def _register_op_flags_and_props(self):
        """Register operating flags and properties.

        Not running the Super method since the i3 properties are sufficiently different than the legacy properties.
        """
        self._register_default_op_flags_and_props(dimmable=True, additional_flags=[])

        self._add_operating_flag(
            TRIGGER_GROUP_MASK, 8, None, None, None, False, PropertyType.ADVANCED
        )
        self._operating_flags[TRIGGER_GROUP_MASK].is_read_only = False
        self._properties[LOAD_BUTTON_NUMBER] = ExtendedProperty(
            address=self._address,
            name=LOAD_BUTTON_NUMBER,
            value_type=int,
            is_reversed=False,
            is_read_only=False,
            prop_type=PropertyType.ADVANCED,
        )
        self._properties[LOAD_BUTTON_NUMBER].set_value(1)
        self._add_ext_prop_read_manager({})
        self._add_ext_prop_write_manager(
            properties={3: self._properties[LOAD_BUTTON_NUMBER]}, data2=0x13
        )

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )
        self._config[RADIO_BUTTON_GROUPS] = RadioButtonGroupsProperty(
            self, RADIO_BUTTON_GROUPS
        )
        for group in self._groups:
            if group == 1:
                continue
            button = self._buttons[group]
            name = f"{TOGGLE_BUTTON}_{button[-1]}"
            self._config[name] = ToggleButtonProperty(
                self._address,
                name,
                group,
                self.properties[NON_TOGGLE_MASK],
                self.properties[NON_TOGGLE_ON_OFF_MASK],
            )
        self._config[LOAD_BUTTON] = LoadButtonProperty(
            self._address,
            LOAD_BUTTON,
            self._properties[LOAD_BUTTON_NUMBER],
            list(self._groups),
        )

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[SET_LEDS_COMMAND] = SetLedsCommandHandler(address=self.address)
        self._add_ext_prop_write_manager(
            {3: self._operating_flags[TRIGGER_GROUP_MASK]}, 0x0C, 0x00, 0x00
        )

    def _subscribe_to_handelers_and_managers(self):
        # Not running super due to the LOAD_BUTTON_NUMBER config property
        super()._subscribe_to_handelers_and_managers()
        self._managers[STATUS_COMMAND].add_status_type(1, self._async_handle_led_status)
        for group in self._buttons:
            led_method = partial(self._async_led_follow_check, group=group)
            self._managers[group][ON_LEVEL_MANAGER].subscribe(
                led_method,
            )
        self._properties[LOAD_BUTTON_NUMBER].subscribe(
            self._handle_load_button_property__changed
        )

    def _handle_status(self, db_version, status):
        """Handle the status response."""
        # Need to check which button is the load button
        load = self.load_button
        relay_mode = self._operating_flags[RELAY_MODE_OFF].value
        relay_mode = False if relay_mode is None else relay_mode
        self._groups[load].is_dimmable = not relay_mode
        self._groups[load].set_value(status)

    async def _async_handle_led_status(self, db_version, status):
        """Set the on level of the LED from a status command."""
        load_button = self._properties[LOAD_BUTTON_NUMBER].value
        load_button = 1 if load_button in [0, None] else load_button
        for bit in range(1, 9):
            state = self._groups.get(bit)
            if state:
                value = bit_is_set(status, bit - 1)
                if bit == self.load_button and bool(state.value) != value:
                    await self.async_status()
                else:
                    state.is_dimmable = False
                    state.value = value

    def _get_led_follow_state(self, button1, button2, new_state) -> bool:
        """Return the state of button2 based on new_state of button1."""
        if button1 == button2:
            return new_state
        button2_str = f"_{button2}" if button2 != 1 else ""
        on_mask = self._properties[f"{ON_MASK}{button2_str}"]
        off_mask = self._properties[f"{OFF_MASK}{button2_str}"]
        if not on_mask.is_loaded or not off_mask.is_loaded:
            return False
        follow = bit_is_set(on_mask.value, button1 - 1)
        set_off = bit_is_set(off_mask.value, button1 - 1)
        if follow:
            # If button1 on and off mask set then turn off button2
            if set_off and new_state:
                return False
            # Else we follow the state as is?????
            return bool(new_state)
        return bool(self._groups[button2].value)

    def _has_follow_leds(self, button):
        """Check if the current button has follow buttons."""
        for follow_button in self._groups:
            if follow_button == button:
                continue
            button_str = f"_{follow_button}" if follow_button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            if not on_mask.is_loaded:
                return False
            follow = bit_is_set(on_mask.value, button - 1)
            if follow:
                return True
        return False

    def _calc_set_leds_command_kwargs(self, led, is_on) -> Dict[str, bool]:
        """Calculate the kwargs values for the SetLedsCommandHandler."""
        leds = {}
        for curr_led in range(1, 9):
            name = f"group{curr_led}"
            if curr_led == led:
                leds[name] = is_on
            else:
                curr_group = self._groups.get(curr_led)
                if curr_group:
                    led_val = self._get_led_follow_state(led, curr_led, is_on)
                else:
                    led_val = False
                leds[name] = led_val
        return leds

    async def _async_update_led_state(
        self,
        group: int,
        value: Union[int, bool],
        event: str,
        led_kwargs: Dict[str, bool],
    ):
        """Set the value of an LED based on the set state and the configuration."""
        if not self._groups.get(group):
            return
        if self._properties[NON_TOGGLE_MASK].value is not None:
            non_toogle = bit_is_set(self._properties[NON_TOGGLE_MASK].value, group)
        else:
            non_toogle = False
        if non_toogle:
            self._groups[group].value = 0
        else:
            self._groups[group].value = value
        self._events[group][event].trigger(value)

        for name, btn_value in led_kwargs.items():
            button = int(name[-1])
            state = self._groups.get(button)
            if state:
                if button == self.load_button and bool(state.value) != btn_value:
                    await self.async_status()
                else:
                    state.value = btn_value

    async def _async_led_follow_check(self, group, on_level):
        """Update LEDs that follow the changed LED."""
        for curr_group, state in self._groups.items():
            if curr_group == group:
                continue
            follow_state = self._get_led_follow_state(group, curr_group, on_level)
            if curr_group == self.load_button and bool(state.value) != follow_state:
                await self.async_status()
            else:
                state.value = follow_state

    def _handle_load_button_property__changed(self, name, value):
        """Handle load button changes."""
        orig_load_button = 1
        for group, state in self._groups.items():
            if state.is_dimmable:
                orig_load_button = group
            if group == value:
                relay_mode = self._operating_flags[RELAY_MODE_OFF].value
                relay_mode = False if relay_mode is None else relay_mode
                state.is_dimmable = not relay_mode
            else:
                state.is_dimmable = False
        load_value = self._groups[orig_load_button].value
        # Set the orig load button on or off
        self._groups[orig_load_button].value = load_value
        self._groups[value].value = load_value



"""Low battery manager."""

import asyncio

from ..address import Address
from ..handlers.from_device.off import OffInbound
from ..handlers.from_device.on_level import OnLevelInbound
from ..subscriber_base import SubscriberBase
from ..utils import subscribe_topic

WAIT_TIME = 5


class LowBatteryManager(SubscriberBase):
    """Low battery manager."""

    class LowBatterySubscriber(SubscriberBase):
        """Low battery event subscriptions."""

        def call_subscribers(self, low_battery):
            """Call subscribers of this event."""
            self._call_subscribers(low_battery=low_battery)

    def __init__(self, address, group):
        """Init the LowBatteryManager class."""
        self._address = Address(address)
        self._group = group
        subscriber_topic = f"subscriber_{self._address.id}_low_battery"
        super().__init__(subscriber_topic)

        self._on_low_battery = OnLevelInbound(self._address, self._group)
        self._off_low_battery = OffInbound(self._address, self._group)
        self._on_low_battery.subscribe(self._low_battery)
        self._off_low_battery.subscribe(self._low_battery)
        self._low_battery_recd = False
        self._low_battery_state = False
        self._low_battery_event = self.LowBatterySubscriber(f"{subscriber_topic}.true")
        self._low_battery_clear_event = self.LowBatterySubscriber(
            f"{subscriber_topic}.false"
        )
        subscribe_topic(self._all_device_messages, self._address.id)

    def subscribe_low_battery_event(self, callback):
        """Subscribe to low battery event."""
        self._low_battery_event.subscribe(callback)

    def subscribe_low_battery_clear_event(self, callback):
        """Subscribe to low battery event."""
        self._low_battery_clear_event.subscribe(callback)

    def _all_device_messages(self, **kwargs):
        """Capture all messages for this device."""
        loop = asyncio.get_event_loop()
        target = kwargs.get("target")
        # stop if this is a low battery message
        if target and Address(target).low == self._group:
            return
        self._low_battery_recd = False
        loop.call_later(WAIT_TIME, self._check_low_battery)

    def _low_battery(self, on_level):
        """Low battery message received."""
        self._low_battery_recd = True
        self._low_battery_state = True
        self._low_battery_event.call_subscribers(low_battery=True)

    def _check_low_battery(self):
        """Check if a low battery message was received since the last message."""
        if not self._low_battery_recd and self._low_battery_state:
            self._low_battery_clear_event.call_subscribers(low_battery=False)



"""Manage outbound OFF All-Link Broadcast command to a device."""

from ...topics import OFF
from .all_link_cleanup_command import AllLinkCleanupCommandHandlerBase


class OffAllLinkCleanupCommand(AllLinkCleanupCommandHandlerBase):
    """Manage outbound OFF All-Link Broadcast command to a device."""

    def __init__(self, address, group):
        """Init the OnLevelCommand class."""
        super().__init__(topic=OFF, address=address, group=group)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the ON command async."""
        return await super().async_send(group=0, cmd2=self._group)



"""Handle sending a read request for ALDB records."""
import logging

from .. import ack_handler
from ...address import Address
from ...topics import EXTENDED_GET_SET
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class ExtendedSetCommand(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address, data1=None, data2=None, cmd2=0x00):
        """Init the ExtendedSetCommand."""
        if data2 in [0, 1]:
            _LOGGER.error("Extended Set sent with bad action number: %d", data1)
            raise ValueError("Error creating extended set command")
        super().__init__(topic=EXTENDED_GET_SET, address=address, group=0)
        self._cmd2 = cmd2
        self._data1 = data1
        self._data2 = data2

    # pylint: disable=arguments-differ
    async def async_send(
        self,
        data2=None,
        data3=0,
        data4=0,
        data5=0,
        data6=0,
        data7=0,
        data8=0,
        data9=0,
        data10=0,
        data11=0,
        data12=0,
        data13=0,
        data14=0,
        priority=5,
    ):
        """Send Get Operating Flags message asyncronously."""
        data2 = data2 if data2 is not None else self._data2
        kwargs = {"data1": self._data1, "data2": data2}
        loc = locals()
        for item in range(3, 15):
            kwargs[f"data{item}"] = int(loc[f"data{item}"])
        return await super().async_send(cmd2=self._cmd2, priority=5, **kwargs)

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle the ACK."""
        if (
            cmd2 == self._cmd2
            and (self._data1 is None or user_data["d1"] == self._data1)
            and (self._data2 is None or user_data["d2"] == self._data2)
        ):
            await super().async_handle_ack(cmd1=cmd1, cmd2=cmd2, user_data=user_data)



"""Window Covering devices."""
from ..config import (
    DUAL_LINE_ON,
    DURATION_HIGH,
    DURATION_LOW,
    FORWARD_ON,
    KEY_BEEP_ON,
    LED_BLINK_ON_ERROR_OFF,
    LED_BLINK_ON_TX_ON,
    LED_OFF,
    MOMENTARY_LINE_ON,
    NOT_3_WAY,
    ON_LEVEL,
    PROGRAM_LOCK_ON,
    RAMP_RATE,
    RAMP_RATE_IN_SEC,
    X10_HOUSE,
    X10_UNIT,
)
from ..config.ramp_rate import RampRateProperty
from ..constants import PropertyType
from ..groups import COVER
from .open_close_responder_base import OpenCloseResponderBase


class WindowCovering(OpenCloseResponderBase):
    """Window Covering device."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the WindowCovering class."""
        super().__init__(
            address, cat, subcat, firmware, description, model, state_name=COVER
        )

    def _register_op_flags_and_props(self):
        """Register the operating and properties."""
        super()._register_op_flags_and_props()

        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_OFF, 0, 4, 0x0A, 0x0B)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0C, 0x0D)

        self._add_operating_flag(LED_BLINK_ON_ERROR_OFF, 2, 3, 0x15, 0x16)

        self._add_operating_flag(DUAL_LINE_ON, 3, 0, 0x1E, 0x1F)
        self._add_operating_flag(MOMENTARY_LINE_ON, 3, 1, 0x20, 0x21)
        self._add_operating_flag(NOT_3_WAY, 3, 3, 0x22, 0x23)
        self._add_operating_flag(FORWARD_ON, 3, 4, 0x24, 0x25)

        self._add_property(X10_HOUSE, 5, None, prop_type=PropertyType.ADVANCED)  # 4
        self._add_property(X10_UNIT, 6, None, prop_type=PropertyType.ADVANCED)  # 4
        self._add_property(RAMP_RATE, 7, 5, prop_type=PropertyType.ADVANCED)

        # Need to verify use_data position
        self._add_property(ON_LEVEL, 8, 6)
        self._add_property(DURATION_HIGH, 9, None)  # 0x10
        self._add_property(DURATION_LOW, 10, None)  # 0x10

    def _register_config(self):
        """Register configuration items."""
        super()._register_config()
        self._config[RAMP_RATE_IN_SEC] = RampRateProperty(
            self._address, RAMP_RATE_IN_SEC, self._properties[RAMP_RATE]
        )



"""Convert an Insteon Message to a URL for the Hub."""
from ..constants import AllLinkMode
from .messages.outbound import Outbound


def _to_url(host, port, cmd):
    """Convert a host, port and command to a url."""
    return f"http://{host}:{port}/{cmd}"


def convert_to_url(host, port, msg: Outbound) -> str:
    """Convert a message to a URL."""
    return _to_url(host, port, f"3?{bytes(msg).hex()}=I=3")


def start_all_link_url(host, port, msg: Outbound):
    """Convert a Start All-Linking message to URL."""
    if msg.link_mode == AllLinkMode.DELETE:
        return _to_url(host, port, f"3?{bytes(msg).hex()}=I=3")
    return _to_url(host, port, f"0?09{msg.group:d}=I=0")


def cancel_all_linking_url(host, port, msg: Outbound):
    """Convert a Cancel All-Linking message to URL."""
    return _to_url(host, port, "0?08=I=0")



"""Manage outbound Png command to a device."""
from ...topics import PING
from .direct_command import DirectCommandHandlerBase


class PingCommand(DirectCommandHandlerBase):
    """Manage an outbound Ping command to a device."""

    def __init__(self, address):
        """Init the PingCommand class."""
        super().__init__(topic=PING, address=address)

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers()



"""Test X10 address."""
import unittest

from pyinsteon.x10_address import X10Address, create
from tests import set_log_levels


class TestX10Address(unittest.TestCase):
    """Test X10 address."""

    def setUp(self):
        """Set up the test."""
        self.housecode = "C"
        self.housecode_byte = 0x02
        self.unitcode = 6
        self.unitcode_byte = 0x09
        self.address = X10Address((self.housecode_byte << 4) + self.unitcode_byte)
        self.address_create = create(self.housecode, self.unitcode)
        set_log_levels(
            logger="info",
            logger_pyinsteon="info",
            logger_messages="info",
            logger_topics=False,
        )

    def test_bytes(self):
        """Test byte conversion."""
        byte_out = bytes(bytearray([(self.housecode_byte << 4) + self.unitcode_byte]))
        assert bytes(self.address) == byte_out

    def test_id(self):
        """Test the X10 address id."""
        assert self.address.id == "x10c06"

    def test_string(self):
        """Test the X10 address as string."""
        assert str(self.address) == "X10.C.06"

    def test_getitem(self):
        """Test get houscode and unit code."""
        assert self.address[0] == self.housecode_byte
        assert self.address[1] == self.unitcode_byte
        try:
            # pylint: disable=unused-variable
            fail_test = self.address[2]  # noqa: F841
            assert False
        except ValueError:
            assert True

    def test_eq(self):
        """Test equals."""
        assert self.address == X10Address((0x02 << 4) + 0x09)

    def test_ne(self):
        """Test not equals."""
        assert self.address != X10Address((0x01 << 4) + 0x09)

    def test_create(self):
        """Test create from unitcode and housecode."""
        test_bytes = bytes(bytearray([(self.housecode_byte << 4) + self.unitcode_byte]))
        assert bytes(self.address_create) == test_bytes

    def test_from_string(self):
        """Test create from string."""
        addr = X10Address("X10.A.02")
        assert bytes(addr) == b"n"

    def test_for_all_units_lights_on_off(self):
        """Test creating an X10 address for use in the All Lights and All Units on off commands."""
        addr = create("a", 0)
        assert isinstance(addr, X10Address)


if __name__ == "__main__":
    unittest.main()



"""Base class to handle Broadcast messages from devices."""

from datetime import datetime

from .. import broadcast_handler
from ...constants import MessageFlagType
from ..inbound_base import InboundHandlerBase

MIN_DUP = 0.7
MAX_DUP = 2


class BroadcastCommandHandlerBase(InboundHandlerBase):
    """Base class to handle inbound Broadcast messages."""

    def __init__(self, topic, address, group=None):
        """Init the broadcast_handlerBase class."""
        self._group = group
        super().__init__(
            topic=topic,
            address=address,
            group=group,
            message_type=MessageFlagType.ALL_LINK_BROADCAST,
        )
        self._last_command = datetime(1, 1, 1)
        self._last_hops_left = None
        self._last_cmd2 = None
        self._last_userdata = None

    @broadcast_handler
    def receive_message(self, cmd1, cmd2, target, user_data, hops_left):
        """Receive the inbound message."""
        if not self._is_first_message(cmd2, user_data, target, hops_left):
            return
        self._handle_message_received(cmd1, cmd2, target, user_data, hops_left)

    def _handle_message_received(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the received message."""

    def _is_first_message(self, cmd2, user_data, target, hops_left):
        """Test if the message is a duplicate."""
        curr_time = datetime.now()
        tdelta = curr_time - self._last_command
        if tdelta.days > 0:
            delta = 9999
        else:
            delta = tdelta.seconds + tdelta.microseconds / 1000000
        self._last_command = curr_time

        if cmd2 != self._last_cmd2 or user_data != self._last_userdata:
            self._last_cmd2 = cmd2
            self._last_userdata = user_data
            self._last_hops_left = hops_left
            return True
        self._last_cmd2 = cmd2
        self._last_userdata = user_data

        if target.middle != 0 and hops_left == self._last_hops_left and delta < MIN_DUP:
            return False

        if (
            self._last_hops_left is None
            or (hops_left == self._last_hops_left and delta > MIN_DUP)
            or hops_left > self._last_hops_left
            or delta >= MAX_DUP
        ):
            self._last_hops_left = hops_left
            return True
        return False
